#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        SemiDevOpReader.py
# Purpose:     Read Semiconductor Device Operating Points from a log file
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     19-09-2021
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------

"""
Implements a parser for extracting Semiconductor Devices Operating Points from an LTSpice log file.
"""

import re
from ..utils.detect_encoding import detect_encoding


def opLogReader(filename: str) -> dict:
    """
    This function is exclusively dedicated to retrieving operation point parameters of Semiconductor Devices. This is
    handled separately from the main LogReader class because of its specialization and therefore not judged to be
    of interest to the typical LTSpice user making board level simulations.

    Below is an excerpt of a Semiconductor Device Operating Points log.py

    .. code-block:: text

        Semiconductor Device Operating Points:

                                --- Diodes ---
        Name:         d:m6:1:para2:1               d:m4:1:para2:1               d:m3:1:para2:1         d:m6:1:para1:2
        Model: m6:1:para2:dpar_5v_psubnburd m4:1:para2:dpar_5v_psubnburd m3:1:para2:dpar_5v_psubnburd dpar_pburdnburdsw
        Id:              3.45e-19                     3.45e-19                     3.45e-19                1.11e-13
        Vd:              3.27e-07                     3.27e-07                     3.27e-07                9.27e-02
        ...
        CAP:             3.15e-14                     3.15e-14                     3.15e-14                5.20e-14

                            --- Bipolar Transistors ---
        Name:     q:q2:1:2    q:q2:1:1     q:q1:1:2    q:q1:1:1     q:q7:3
        Model:  q2:1:qnl_pnp q2:1:qnl_m  q1:1:qnl_pnp q1:1:qnl_m   qpl_parc
        Ib:       3.94e-12     4.69e-08    7.43e-13     4.70e-08    3.75e-12
        Ic:      -2.34e-12     4.57e-06   -7.44e-13     4.50e-06   -2.35e-12
        Vbe:      1.60e+00     7.40e-01   -7.88e-04     7.40e-01    1.40e+00


    This function will parse the log file and will produce a dictionary that contains all the information retrieved with
    the following format:

    .. code-block:: python

        semi_ops = {
            'Diodes': {
                'd:m6:1:para2:1': {
                    'Model': 'm6:1:para2:dpar_5v_psubnburd', 'Id': 3.45e-19, 'Vd': 3.27e-07, ... 'CAP': 3.15e-14 },
                'd:m4:1:para2:1': {
                    'Model': 'm4:1:para2:dpar_5v_psubnburd', 'Id': 3.45e-19, 'Vd': 3.27e-07, ..., 'CAP': 3.15e-14 },
                'd:m3:1:para2:1': {
                    'Model': 'm3:1:para2:dpar_5v_psubnburd', 'Id': 3.45e-19, 'Vd': 3.27e-07, ..., 'CAP': 3.15e-14 },
                'd:m6:1:para1:2': {
                    'Model': 'dpar_pburdnburdsw', 'Id': 1.11e-13, 'Vd': 0.0927, ..., 'CAP': 5.2e-14 },
            },
            'Bipolar Transistors': {
                'q:q2:1:2': {
                    'Model': 'q2:1:qnl_pnp', 'Ib': 3.94e-12, 'Ic': -2.34e-12, 'Vbe': 160, ... },
                'q:q2:1:1': {
                    'Model': 'q2:1:qnl_m', 'Ib': 4.69e-08, 'Ic': 4.57e-06, 'Vbe': 0.074, ... },
                'q:q1:1:2': {
                    'Model': 'q1:1:qnl_pnp', 'Ib': 7.43e-13, 'Ic': -7.44e-13, 'Vbe': -0.000788, ... },
                'q:q1:1:1': {
                    'Model': 'q1:1:qnl_m', 'Ib': 4.7e-08, 'Ic': 4.5e-06, 'Vbe': 0.74, ... },
                'q:q7:3': {
                    'Model': 'qpl_parc', 'Ib': 3.75e-12, 'Ic': -2.35e-12, 'Vbe': 1.4, ... },
            },
        }


    :param filename: path to the log file containing the Semiconductor Device Operating Points
    :type filename: str
    :return: Dictionary containing the information as described above.
    :rtype: dict
    """
    dataset = {}
    is_title = re.compile(r"^\s*--- (.*) ---\s*$")
    encoding = detect_encoding(filename)
    log = open(filename, 'r', encoding=encoding)
    where = None
    n_devices = 0
    line = None
    for line in log:
        if line.startswith("Semiconductor Device Operating Points:"):
            break

    if line is not None and line.startswith("Semiconductor Device Operating Points:"):
        for line in log:
            match = is_title.search(line)
            if match is not None:
                where = match.group(1)
                dataset[where] = {}  # Creates a dictionary for each component type
            else:
                cols = re.split(r'\s+', line.rstrip('\r\n'))
                if len(cols) > 1 and (cols[0].endswith(":") or cols[0] == 'Gmb'):  # The last 'or condition solves an
                    # LTSpice bug where the Gmb parameter is not suffixed by :  - Thanks to Amitkumar for finding this.
                    if cols[0] == "Name:":
                        devices = cols[1:]
                        n_devices = len(devices)
                        for dev in cols[1:]:
                            dataset[where][dev] = {}
                    else:
                        if n_devices > 0 and len(cols) == (n_devices + 1):
                            param = cols[0].rstrip(':')
                            for i, val in enumerate(cols[1:]):
                                try:
                                    value = float(val)
                                except ValueError:
                                    value = val
                                dataset[where][devices[i]][param] = value
    log.close()
    return dataset



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        sim_analysis.py
# Purpose:     Classes to automate Monte-Carlo, FMEA or Worst Case Analysis
#              be updated by user instructions
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     06-07-2021
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------

from collections import OrderedDict
from copy import deepcopy
from functools import wraps
from pathlib import Path
from typing import Union, Optional, Type, Callable
import logging
from ..sim_runner import AnyRunner, RunTask, ProcessCallback
from ...editor.base_editor import BaseEditor
from ...log.logfile_data import LogfileData
from ...utils.detect_encoding import EncodingDetectError

from ...log.ltsteps import LTSpiceLogReader
from ...log.qspice_log_reader import QspiceLogReader

_logger = logging.getLogger("spicelib.SimAnalysis")

class SimAnalysis(object):
    """
    Base class for making Monte-Carlo, Extreme Value Analysis (EVA) or Failure Mode and Effects Analysis.
    As a base class, a certain number of assertions must be made on the simulation results that will make the pass/fail.

    Note: For the time being only measurements done with .MEAS are possible. At a later stage the parsing of RAW files
    will be possible, although, it seems that the later solution is less computing intense.
    """

    def __init__(self, circuit_file: Union[str, BaseEditor], runner: Optional[AnyRunner] = None):
        if isinstance(circuit_file, str):
            from ...editor.spice_editor import SpiceEditor
            self.editor = SpiceEditor(circuit_file)
        else:
            self.editor = circuit_file
        self._runner = runner
        self.simulations = []
        self.last_run_number = 0
        self.received_instructions = []
        self.instructions_added = False
        self.log_data = LogfileData()

    def clear_simulation_data(self):
        """Clears the data from the simulations"""
        self.simulations.clear()

    @property
    def runner(self):
        if self._runner is None:
            from ...sim.sim_runner import SimRunner
            self._runner = SimRunner()
        return self._runner

    @runner.setter
    def runner(self, new_runner: AnyRunner):
        self._runner = new_runner

    def run(self, *,
            wait_resource: bool = True,
            callback: Union[Type[ProcessCallback], Callable] = None,
            callback_args: Union[tuple, dict] = None,
            switches=None,
            timeout: float = None, run_filename: str = None) -> Union[RunTask, None]:
        """
        Runs the simulations. See runner.run() method for details on arguments.
        """
        sim = self.runner.run(self.editor,
                              wait_resource=wait_resource,
                              callback=callback,
                              callback_args=callback_args,
                              switches=switches,
                              timeout=timeout,
                              run_filename=run_filename)
        if sim is not None:
            self.simulations.append(sim)
            return sim

    def wait_completion(self):
        self.runner.wait_completion()

    @wraps(BaseEditor.reset_netlist)
    def reset_netlist(self):
        """Resets the netlist to the original state and clears the instructions added by the user."""
        self._reset_netlist()
        self.received_instructions.clear()

    def _reset_netlist(self):
        """Unlike the reset_netlist method of the BaseEditor, this method does not clear the instructions added by the
        user. This is useful for the case where the user wants to run multiple simulations with different parameters
        without having to add the instructions again."""
        self.editor.reset_netlist()
        self.instructions_added = False

    def set_component_value(self, ref: str, new_value: str):
        self.received_instructions.append(('set_component_value', ref, new_value))

    def set_element_model(self, ref: str, new_model: str):
        self.received_instructions.append(('set_element_model', ref, new_model))

    def set_parameter(self, ref: str, new_value: str):
        self.received_instructions.append(('set_parameter', ref, new_value))

    def add_instruction(self, new_instruction: str):
        self.received_instructions.append(('add_instruction', new_instruction))

    def remove_instruction(self, instruction: str):
        self.received_instructions.append(('remove_instruction', instruction))

    def remove_Xinstruction(self, search_pattern: str):
        self.received_instructions.append(('remove_Xinstruction', search_pattern))

    def play_instructions(self):
        if self.instructions_added:
            return  # Nothing to do
        for instruction in self.received_instructions:
            if instruction[0] == 'set_component_value':
                self.editor.set_component_value(instruction[1], instruction[2])
            elif instruction[0] == 'set_element_model':
                self.editor.set_element_model(instruction[1], instruction[2])
            elif instruction[0] == 'set_parameter':
                self.editor.set_parameter(instruction[1], instruction[2])
            elif instruction[0] == 'add_instruction':
                self.editor.add_instruction(instruction[1])
            elif instruction[0] == 'remove_instruction':
                self.editor.remove_instruction(instruction[1])
            elif instruction[0] == 'remove_Xinstruction':
                self.editor.remove_Xinstruction(instruction[1])
            else:
                raise ValueError("Unknown instruction")
        self.instructions_added = True

    def save_netlist(self, filename: str):
        self.play_instructions()
        self.editor.save_netlist(filename)

    def cleanup_files(self):
        """Clears all simulation files. Typically used after a simulation run and analysis."""
        self.runner.file_cleanup()

    def simulation(self, index: int):
        """Returns a simulation object"""
        return self.simulations[index]

    def __getitem__(self, item):
        return self.simulations[item]

    @staticmethod
    def read_logfile(run_task: RunTask) -> Union[LogfileData, None]:
        """Reads the log file and returns a dictionary with the results"""
        if run_task.simulator.__name__ == "LTspice":
            LogReader = LTSpiceLogReader
        elif run_task.simulator.__name__ == "Qspice":
            LogReader = QspiceLogReader
        else:
            raise ValueError("Unknown simulator type")

        try:
            log_results = LogReader(run_task.log_file)
        except FileNotFoundError:
            _logger.warning("Log file not found: %s", run_task.log_file)
            return None
        except EncodingDetectError:
            _logger.warning("Log file %s couldn't be read", run_task.log_file)
            return None
        return log_results

    def add_log_data(self, log_data: LogfileData):
        """Add data from a log file to the log_data object"""
        if log_data is None:
            return

        for param in log_data.stepset:
            if param not in self.log_data.stepset:
                self.log_data.stepset[param] = log_data.stepset[param]
            else:
                self.log_data.stepset[param].extend(log_data.stepset[param])
        for param in log_data.dataset:
            if param not in self.log_data.dataset:
                self.log_data.dataset[param] = log_data.dataset[param][:]
            else:
                self.log_data.dataset[param].extend(log_data.dataset[param][:])
        self.log_data.step_count += log_data.step_count

    def read_logfiles(self) -> LogfileData:
        """Reads the log files and returns a dictionary with the results"""
        self.log_data = LogfileData()  # Clears the log data
        for sim in self.simulations:
            if sim is None:
                continue

            log_results = self.read_logfile(sim)
            self.add_log_data(log_results)

        return self.log_data

    def configure_measurement(self, meas_name: str, meas_expression: str, meas_type: str = 'tran'):
        """Configures a measurement to be done in the simulation"""
        self.editor.add_instruction(".meas {} {} {}".format(meas_type, meas_name, meas_expression))



#!/usr/bin/env python
# coding=utf-8

"""Helper script to be able to insert validated code into the README.md file."""
import os
import sys
import re
from shutil import copyfile

# Read the README.md on the path indicated on the command line arguments
filename = sys.argv[1]
print(f"Reading \"{filename}\"", end="...")
try:
    readme_md = open(filename).readlines()
except FileNotFoundError:
    print("File not found")
    exit(1)
except Exception as e:
    print(f"Error reading file: {e}")
    exit(1)
print(f"{len(readme_md)} lines read")
in_statement_regex = re.compile(r"-- in (?P<path>.*?)(?P<loc>\s\[.*\])?\s*$")

block_start = -1
line_no = 0
# detect python code blocks
while line_no < len(readme_md):
    line = readme_md[line_no]
    if "```python" in line:
        block_start = line_no
    elif "```" in line and block_start != -1:
        # check whether the next line has the -- in statement
        if line_no + 1 < len(readme_md):
            m = in_statement_regex.search(readme_md[line_no + 1])
            if not m and line_no + 2 < len(readme_md):  # if there is no -- in statement, check the next line
                m = in_statement_regex.search(readme_md[line_no + 2])
            # find the file to include
            if m:
                print(f"Updating code on lines {block_start + 1}:{line_no + 1}")
                include_relpath = m.group("path")
                include_path = os.path.abspath(os.path.join(os.curdir, include_relpath))
                include_text = open(include_path, "r", encoding="utf-8").readlines()
                # if there is a localization, isolate only the part identified by loc
                # the part starts with "# -- Start of <loc> --" and ends with "# -- End of <loc> --"
                # where <loc> is the text within the square brackets [] in the readme.md
                if (loc := m.group("loc")) is not None:
                    lines_to_plug = []
                    loc_text = loc[2:-1]  # This is crude but it works. It removes the leading \s[ and the trailing ]
                    start_tag = "-- Start of %s --" % loc_text
                    end_tag = "-- End of %s --" % loc_text
                    include_ident = -1
                    for line1 in include_text:
                        if start_tag in line1:
                            include_ident = line1.find("#")  # This marks the indent of all the block
                        elif end_tag in line1:
                            include_ident = -1
                        else:
                            if include_ident >= 0:  # means that we are reading between the Start and End
                                if include_ident >= len(line1):  # if it is a blank line
                                    lines_to_plug.append("")  # Add a blank line
                                else:
                                    lines_to_plug.append(line1[include_ident:])
                else:
                    lines_to_plug = include_text
                existing_lines = line_no - (block_start + 1)  # This accounts for the start and finish ``` line
                new_lines = len(lines_to_plug)
                readme_md[block_start + 1:line_no] = lines_to_plug
                line_no += new_lines - existing_lines
        block_start = -1
    line_no += 1

# Finally write back the readme.md but first create a backup
backup_filename = filename.replace(".md", ".bak")
print(f"Creating backup {backup_filename}")
copyfile(filename, backup_filename)
print(f"Writing {len(readme_md)} lines to {filename}")
open(filename, 'w').writelines(readme_md)
exit(0)



#!/usr/bin/env python
# coding=utf-8
# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        fast_worst_case.py
# Purpose:     Class to automate Worst-Case simulations (Faster Algorithm)
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     04-11-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------

import logging
from time import sleep
from typing import Union, Callable, Type
from enum import IntEnum

from .worst_case import WorstCaseAnalysis, DeviationType, ToleranceDeviations
from ..process_callback import ProcessCallback
from ...log.logfile_data import LogfileData

_logger = logging.getLogger("spicelib.SimAnalysis")


class WorstCaseType(IntEnum):
    nom = 0
    max = 1
    min = 2


class FastWorstCaseAnalysis(WorstCaseAnalysis):
    """
    This class implements a faster algorithm to perform a worst case analysis. The typical worst case analysis makes
    all possible combinations of the components that have a deviation. This means that if there are 10 components with
    a deviation, there will be 1024 simulations to be performed. The number of simulations grows exponentially with the
    number of components with deviation.

    This algorithm speeds up the process, by determining the impact of each component with deviation on the final
    result and skipping simulations that are not necessary. The algorithm is as follows:

        1. Make a sensitivity analysis to determine the impact that each component has on the final result.

        2. Based on the information collected on 1. set all the components with deviation achieve a maximum
        on the final result. Components are set based on the assumption that the system is reaction is to each
        component is monotonic. That is, if increasing the component value, increases the final result, then the
        component is set to the maximum value. If decreasing the component value increases the final result, then the
        component is set to the minimum value.

        3. Validate the assumption based on 2. by making a series of test simulations setting each component with their
        opposite value. If the result is lower, then the assumption is correct. If the result is higher, then the
        assumption is wrong and the component is set to the opposite value.

        4. Repeat 2. but this time trying to achieve a minimum on the final result.

        5. Repeat step 3. but this time trying to achieve a minimum on the final result.

    Like in the Worst-Case and Montecarlo, there are two approaches to make this analysis. Either preparing a testbench
    where component variations are managed by the simulator, or by ordering each simulation individually.
    In the testbench method all component values are replaced by formulas that depend on a .STEP PARAM run, and
    then run_testbench() will make all the manipulations of the run variable.

    In the latter, each component value is set individually. This is done by calling the run_analysis() method.
    """

    def run_testbench(
            self, *,
            runs_per_sim: int = None,  # This parameter is ignored
            wait_resource: bool = True, # This parameter is ignored
            callback: Union[Type[ProcessCallback], Callable] = None,
            callback_args: Union[tuple, dict] = None,
            switches=None,
            timeout: float = None,
            run_filename: str = None) -> None:
        raise NotImplementedError("run_testbench() is not implemented in this class")

    def run_analysis(self,
                     callback: Union[Type[ProcessCallback], Callable] = None,
                     callback_args: Union[tuple, dict] = None,
                     switches=None,
                     timeout: float = None,
                     measure: str = None,
                     ):
        """
        As described in the class description, this method will perform a worst case analysis using a faster algorithm.
        """
        assert measure is not None, "The measure argument must be defined"

        self.clear_simulation_data()
        self.elements_analysed.clear()
        worst_case_elements = {}

        def check_and_add_component(ref1: str):
            val1, dev1 = self.get_component_value_deviation_type(ref1)  # get there present value
            if dev1.min_val == dev1.max_val or dev1.typ == DeviationType.none:
                return
            worst_case_elements[ref1] = val1, dev1, 'component'
            self.elements_analysed.append(ref1)

        def value_change(val, dev, to: WorstCaseType):
            """
            Sets the reference component to the maximum value if set_max is True, or to the minimum value if set_max is
            False. This method is used by the run_analysis() method.
            """
            # Preparing the variation on components, but only on the ones that have changed
            if dev.typ == DeviationType.tolerance:
                if to == WorstCaseType.max:
                    new_val = val * (1 + dev.max_val)
                elif to == WorstCaseType.min:
                    new_val = val * (1 - dev.max_val)
                else:
                    # Default to nominal case
                    new_val = val

            elif dev.typ == DeviationType.minmax:
                if to == WorstCaseType.max:
                    new_val = dev.max_val
                elif to == WorstCaseType.min:
                    new_val = dev.min_val
                else:
                    # Default to nominal case
                    new_val = val
            else:
                _logger.warning("Unknown deviation type")
                new_val = val
            return new_val

        def set_ref_to(ref, to: WorstCaseType):
            val, dev, typ = worst_case_elements[ref]
            new_val = value_change(val, dev, to)
            if typ == 'component':
                self.editor.set_component_value(ref, new_val)  # update the value
            elif typ == 'parameter':
                self.editor.set_parameter(ref, new_val)
            else:
                _logger.warning("Unknown type")

        def run_and_get_measure():
            # Run the simulation
            task = self.run(
                wait_resource=True,
                callback=callback, callback_args=callback_args,
                switches=switches, timeout=timeout)
            self.wait_completion()
            # Get the results from the simulation
            log_data = self.add_log(task)
            return log_data.get_measure_value(measure)


        for ref in self.device_deviations:
            check_and_add_component(ref)

        for ref in self.parameter_deviations:
            val, dev = self.get_parameter_value_deviation_type(ref)
            if dev.typ == DeviationType.tolerance or dev.typ == DeviationType.minmax:
                worst_case_elements[ref] = val, dev, 'parameter'
                self.elements_analysed.append(ref)

        for prefix in self.default_tolerance:
            for ref in self.get_components(prefix):
                if ref not in self.device_deviations:
                    check_and_add_component(ref)

        _logger.info("Worst Case Analysis: %d elements to be analysed", len(self.elements_analysed))

        self._reset_netlist()  # reset the netlist
        self.play_instructions()  # play the instructions
        # Simulate the nominal case
        self.run(wait_resource=True,
                 callback=callback, callback_args=callback_args,
                 switches=switches, timeout=timeout)

        # Sequence a change of a component value at a time, setting it to the maximum value
        for ref in self.elements_analysed:
            set_ref_to(ref, WorstCaseType.max)
            # Run the simulation
            self.run(wait_resource=True,
                     callback=callback, callback_args=callback_args,
                     switches=switches, timeout=timeout)
        self.wait_completion()
        self.analysis_executed = True  # Need to set this to True, so that the next step can be executed
        self.testbench_executed = True  # Idem
        # Get the results from the simulation
        log_data = self.read_logfiles()
        nominal = log_data.get_measure_value(measure, 0)
        _logger.info("Nominal value: %g", nominal)
        component_deltas = {}
        idx = 1
        new_measure = last_measure = nominal
        for ref in self.elements_analysed:
            new_measure = log_data.get_measure_value(measure, idx)
            component_deltas[ref] = new_measure - last_measure
            last_measure = new_measure
            _logger.info("Component %s: %g", ref, component_deltas[ref])
            idx += 1
        # Check which components have a positive impact on the final result, that is, increasing the component value
        # increases the final result
        max_setting = {ref: component_deltas[ref] > 0 for ref in component_deltas}

        # Set all components with a positive impact to the maximum value and all components with a negative impact to
        # the minimum value
        component_changed = False
        for ref in max_setting:
            if not max_setting[ref]:  # Set the negative impact components to the minimum value
                set_ref_to(ref, WorstCaseType.min)
                component_changed = True

        if component_changed:
            # Run the simulation
            # This is the expected maximum
            max_value = run_and_get_measure()
            idx += 1
        else:
            max_value = new_measure

        # Check if the assumption is correct. Cycling each component to its opposite value
        iterator = iter(self.elements_analysed)
        while True:
            try:
                ref = next(iterator)
            except StopIteration:
                break
            if max_setting[ref]:  # Set the negative impact components to the minimum value
                set_ref_to(ref, WorstCaseType.min)
            else:
                set_ref_to(ref, WorstCaseType.max)
            # Run the simulation
            new_value = run_and_get_measure()
            idx += 1

            if new_value > max_value:
                # The assumption is wrong, so the component is set to the minimum value
                max_setting[ref] = not max_setting[ref]
                max_value = new_value
                # Need to restart the cycle
                iterator = iterator(self.elements_analysed)

            # setting it back to the maximum value
            if max_setting[ref]:
                set_ref_to(ref, WorstCaseType.max)
            else:
                set_ref_to(ref, WorstCaseType.min)

        # Now determining the minimum value: Assuming the opposite of the maximum setting
        min_setting = {ref: not max_setting[ref] for ref in max_setting}

        # Set the component back to their opposite value
        for ref in self.elements_analysed:
            if min_setting[ref]:
                set_ref_to(ref, WorstCaseType.max)
            else:
                set_ref_to(ref, WorstCaseType.min)

        # Run the simulation
        # This is the expected minimum of the final result
        min_value = run_and_get_measure()
        idx += 1

        # Check if the assumption is correct. Cycling each component to its opposite value
        iterator = iter(self.elements_analysed)
        while True:
            try:
                ref = next(iterator)
            except StopIteration:
                break
            if min_setting[ref]:
                set_ref_to(ref, WorstCaseType.min)
            else:
                set_ref_to(ref, WorstCaseType.max)
            # Run the simulation
            new_value = run_and_get_measure()
            idx += 1
            # This is the expected maximum of the final result
            if new_value < min_value:
                # The assumption is wrong, so the component is set to the minimum value
                min_setting[ref] = not min_setting[ref]
                min_value = new_value
                # Need to restart the cycle
                iterator = iter(self.elements_analysed)

            # setting it back to the previous value
            if min_setting[ref]:
                set_ref_to(ref, WorstCaseType.max)
            else:
                set_ref_to(ref, WorstCaseType.min)

        # Now that we have the maximum and minimum values, we can set the components to the nominal value
        min_comp_values = {}
        max_comp_values = {}
        for ref in self.elements_analysed:
            val, dev, typ = worst_case_elements[ref]
            if min_setting[ref]:
                min_comp_values[ref] = value_change(val, dev, WorstCaseType.max)
            else:
                min_comp_values[ref] = value_change(val, dev, WorstCaseType.min)

            if max_setting[ref]:
                max_comp_values[ref] = value_change(val, dev, WorstCaseType.max)
            else:
                max_comp_values[ref] = value_change(val, dev, WorstCaseType.min)

        self.clear_simulation_data()
        self.cleanup_files()
        self.reset_netlist()
        self.play_instructions()

        return nominal, min_value, max_comp_values, max_value, min_comp_values



# coding=utf-8
# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        qsch_editor.py
# Purpose:     Class made to update directly the QSPICE Schematic files
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
import math
import os
import sys
from collections import OrderedDict
from pathlib import Path
from typing import Union, List, Optional, TextIO, Tuple
import re
import logging
from .base_editor import (
    format_eng, ComponentNotFoundError, ParameterNotFoundError,
    PARAM_REGEX, UNIQUE_SIMULATION_DOT_INSTRUCTIONS
)
from .base_schematic import (BaseSchematic, SchematicComponent, Point, ERotation, Line, Text, TextTypeEnum,
                             LineStyle, Shape)
from ..simulators.qspice_simulator import Qspice
from ..utils.file_search import search_file_in_containers

__author__ = "Nuno Canto Brum <nuno.brum@gmail.com>"
__copyright__ = "Copyright 2021, Fribourg Switzerland"

__all__ = ('QschEditor', 'QschTag', 'QschReadingError')

_logger = logging.getLogger("qspice.QschEditor")

QSCH_HEADER = (255, 216, 255, 219)


# «component (-1200,-100) 0 0
QSCH_COMPONENT_POS = 1
QSCH_COMPONENT_ROTATION = 2
QSCH_COMPONENT_ENABLED = 3
#    «symbol V
#       «type: V»
#       «description: Independent Voltage Source»
#       «shorted pins: false»
#       ... primitives : rect, line, zigzag, elipse, etc...
#       «text (100,150) 1 7 0 0x1000000 -1 -1 "R1"»
#       «text (100,-150) 1 7 0 0x1000000 -1 -1 "100K"»
QSCH_SYMBOL_TEXT_REFDES = 0
QSCH_SYMBOL_TEXT_VALUE = 1
#       «pin (0,200) (0,0) 1 0 0 0x0 -1 "+"»
QSCH_SYMBOL_PIN_POS1 = 1
QSCH_SYMBOL_PIN_POS2 = 2
QSCH_SYMBOL_PIN_NET = 8
QSCH_SYMBOL_PIN_NET_BEHAVIORAL = 9
#    »
# »
#   «wire (-1200,100) (-500,100) "N01"»
QSCH_WIRE_POS1 = 1
QSCH_WIRE_POS2 = 2
QSCH_WIRE_NET = 3

# «net (<x>,<y>) <s> <l> <p> "<netname>"»
# (<x>,<y>) - Location of then Net identifier
# <s> - Font Size (1 is default)
# <l> - Location 7=Right 11=Left 13=Bottom 14=Top
#        7 0111
#       11 1011
#       13 1101
#       14 1110
# <p> - 0=Net , 1=Port
QSCH_NET_POS = 1
QSCH_NET_ROTATION = "?"
QSCH_NET_STR_ATTR = 5

#   «text (-800,-650) 1 7 0 0x1000000 -1 -1 "ï»¿.tran 5m"»
QSCH_TEXT_POS = 1
QSCH_TEXT_SIZE = 2
QSCH_TEXT_ROTATION = 3  # 13="0 Degrees" 45="90 Degrees" 77="180 Degrees" 109="270 Degrees" r= 13+32*alpha/90
QSCH_TEXT_COMMENT = 4  # 0="Normal Text" 1="Comment"
QSCH_TEXT_COLOR = 5  # 0xdbbggrr  d=1 "Default" rr=Red gg=Green bb=Blue in hex format
QSCH_TEXT_STR_ATTR = 8

QSCH_TEXT_INSTR_QUALIFIER = "ï»¿"

# «line (2000,1300) (3150,-100) 1 1 0xff0000 -1 -1»
QSCH_LINE_POS1 = 1
QSCH_LINE_POS2 = 2
QSCH_LINE_WIDTH = 3  # 0=Default 1=Thinnest ... 7=Thickest
QSCH_LINE_TYPE = 4  # 0=Normal 1=Dashed 2=Dotted 3=DashDot 4=DashDotDot
QSCH_LINE_COLOR = 5
QSCH_LINE_UNKNOWN1 = 6
QSCH_LINE_UNKNOWN2 = 7

# «rect (1850,1550) (3650,-400) 0 0 0 0x8000 0x1000000 -1 0 -1»
QSCH_RECT_POS1 = 1
QSCH_RECT_POS2 = 2
QSCH_RECT_UNKNOWN0 = 3
QSCH_RECT_LINE_WIDTH = 4
QSCH_RECT_LINE_TYPE = 5  # 0=Normal 1=Dashed 2=Dotted 3=DashDot 4=DashDotDot
QSCH_RECT_LINE_COLOR = 6
QSCH_RECT_FILL_COLOR = 7
QSCH_RECT_UNKNOWN1 = 8
QSCH_RECT_UNKNOWN2 = 9
QSCH_RECT_UNKNOWN3 = 10

#  «ellipse (2100,1150) (2650,150) 0 0 2 0xff0000 0x1000000 -1 -1»
QSCH_ELLIPSE_POS1 = 1
QSCH_ELLIPSE_POS2 = 2
QSCH_ELLIPSE_UNKNOWN0 = 3
QSCH_ELLIPSE_WIDTH = 4
QSCH_ELLIPSE_LINE_TYPE = 5  # 0=Normal 1=Dashed 2=Dotted 3=DashDot 4=DashDotDot
QSCH_ELLIPSE_LINE_COLOR = 6
QSCH_ELLIPSE_FILL_COLOR = 7
QSCH_ELLIPSE_UNKNOWN1 = 8
QSCH_ELLIPSE_UNKNOWN2 = 9

# «arc3p (2700,300) (2250,1200) (2500,800) 0 2 0xff0000 -1 -1»
QSCH_ARC3P_POS1 = 1
QSCH_ARC3P_POS2 = 2
QSCH_ARC3P_POS3 = 3
QSCH_ARC3P_UNKNOWN0 = 4
QSCH_ARC3P_WIDTH = 5
QSCH_ARC3P_LINE_COLOR = 6
QSCH_ARC3P_UNKNOWN1 = 7
QSCH_ARC3P_UNKNOWN2 = 8

# «triangle (3050,1250) (3550,700) (3450,1400) 0 2 0xff0000 0x2000000 -1 -1»
QSCH_TRIANGLE_POS1 = 1
QSCH_TRIANGLE_POS2 = 2
QSCH_TRIANGLE_POS3 = 3
QSCH_TRIANGLE_UNKNOWN0 = 4
QSCH_TRIANGLE_LINE_TYPE = 5  # 0=Normal 1=Dashed 2=Dotted 3=DashDot 4=DashDotDot
QSCH_TRIANGLE_LINE_COLOR = 6
QSCH_TRIANGLE_FILL_COLOR = 7
QSCH_TRIANGLE_UNKNOWN1 = 8
QSCH_TRIANGLE_UNKNOWN2 = 9

# «coil (3050,400) (3450,600) 0 0 2 0xff0000 -1 -1»
QSCH_COIL_POS1 = 1
QSCH_COIL_POS2 = 2
QSCH_COIL_UNKNOWN0 = 3
QSCH_COIL_WIDTH = 4
QSCH_COIL_LINE_TYPE = 5  # 0=Normal 1=Dashed 2=Dotted 3=DashDot 4=DashDotDot
QSCH_COIL_LINE_COLOR = 6
QSCH_COIL_UNKNOWN1 = 7
QSCH_COIL_UNKNOWN2 = 8

# «zigzag (3050,250) (3400,100) 0 0 2 0xff0000 -1 -1»
QSCH_ZIGZAG_POS1 = 1
QSCH_ZIGZAG_POS2 = 2
QSCH_ZIGZAG_UNKNOWN0 = 3
QSCH_ZIGZAG_WIDTH = 4
QSCH_ZIGZAG_LINE_TYPE = 5  # 0=Normal 1=Dashed 2=Dotted 3=DashDot 4=DashDotDot
QSCH_ZIGZAG_LINE_COLOR = 6
QSCH_ZIGZAG_UNKNOWN1 = 7
QSCH_ZIGZAG_UNKNOWN2 = 8


def decap(s: str) -> str:
    """Take the leading < and ending > from the parameter value on a string with the format "param=<value>"
    If they are not there, the string is returned unchanged."""
    regex = re.compile(r"(\w+)=<(.*)>")
    return regex.sub(r"\1=\2", s)


class QschReadingError(IOError):
    ...


class QschTag:
    """
    Class to represent a tag in a QSCH file. It is a recursive class, so it can have children tags.
    """

    def __init__(self, *tokens):
        self.items = []
        self.tokens = []
        if tokens:
            for token in tokens:
                self.tokens.append(str(token))

    @classmethod
    def parse(cls, stream: str, start: int = 0) -> Tuple['QschTag', int]:
        """
        Parses a tag from the stream starting at the given position. The stream should be a string.

        :param stream: The string to be parsed
        :param start: The position to start parsing
        :return: A tuple with the tag and the position after the tag
        """
        self = cls()
        assert stream[start] == '«'
        i = start + 1
        i0 = i
        while i < len(stream):
            if stream[i] == '«':
                child, i = QschTag.parse(stream, i)
                i0 = i + 1
                self.items.append(child)
            elif stream[i] == '»':
                stop = i + 1
                if i > i0:
                    self.tokens.append(stream[i0:i])
                return self, stop
            elif stream[i] == ' ' or stream[i] == '\n':
                if i > i0:
                    self.tokens.append(stream[i0:i])
                i0 = i + 1
            elif stream[i] == '"':
                # get all characters until the next " sign
                i += 1
                while stream[i] != '"':
                    i += 1
            elif stream[i] == '(':
                # todo: support also [] and {}
                nested = 1
                while nested > 0:
                    i += 1
                    if stream[i] == '(':
                        nested += 1
                    elif stream[i] == ')':
                        nested -= 1
            i += 1
        else:
            raise IOError("Missing » when reading file")

    def __str__(self):
        """Returns only the first line of the tag. The children are not shown."""
        return ' '.join(self.tokens)

    def out(self, level):
        """
        Returns a string representation of the tag with the specified indentation.

        :param level: The indentation level
        :return: A string representation of the tag
        """
        spaces = '  ' * level
        if len(self.items):
            return (f"{spaces}«{' '.join(self.tokens)}\n"
                    f"{''.join(tag.out(level + 1) for tag in self.items)}"
                    f"{spaces}»\n")
        else:
            return f"{'  ' * level}«{' '.join(self.tokens)}»\n"

    @property
    def tag(self) -> str:
        """Returns the tag id of the object. The tag id is the first token in the tag."""
        return self.tokens[0]

    def get_items(self, item) -> List['QschTag']:
        """Returns a list of children tags that match the given tag id."""
        answer = [tag for tag in self.items if tag.tag == item]
        return answer

    def get_attr(self, index: int):
        """
        Returns the attribute at the given index. The attribute can be a string, an integer or a tuple.
        The return type depends on the attribute being read.
        If the attribute is between quotes, it returns a string.
        If it is between parenthesis, it returns a tuple of integers.
        If it starts with "0x", it returns an integer representing the following hexadecimal number.
        Otherwise, it returns an integer.

        :param index: The index of the attribute to be read
        :type index: int
        :return: The attribute at the given index
        """
        a = self.tokens[index]
        if a.startswith('(') and a.endswith(')'):
            return tuple(int(x) for x in a[1:-1].split(','))
        elif a.startswith('0x'):
            return int(a[2:], 16)
        elif a.startswith('"') and a.endswith('"'):
            return a[1:-1]
        else:
            try:
                value = int(a)
            except ValueError:
                value = float(a)
            return value

    def set_attr(self, index: int, value):
        """Sets the attribute at the given index. The attribute can be a string, an integer or a tuple.
        Integer values are written as integers, strings are written between quotes unless it starts with "0x"
        and tuples are written between parenthesis.

        :param index: The index of the attribute to be set
        :type index: int
        :param value: The value to be set
        :type value: Union[str, int, tuple]
        :return: Nothing
        """
        if isinstance(value, int):
            value_str = str(value)
        elif isinstance(value, str):
            if value.startswith('0x'):
                value_str = value
            else:
                value_str = f'"{value}"'
        elif isinstance(value, tuple):
            value_str = f'({value[0]},{value[1]})'
        else:
            raise ValueError("Object not supported in set_attr")
        self.tokens[index] = value_str

    def get_text(self, label, default: str = None) -> str:
        """
        Returns the text of the first child tag that matches the given label.
        """
        a = self.get_items(label + ':')
        if len(a) != 1:
            if default is None:
                raise IndexError(f"Label '{label}' not found in:{self}")
            else:
                return default
        return a[0].tokens[1]

    def get_text_attr(self, index: int) -> str:
        """Returns the text of the attribute at the given index. Unlike get_attr, this method only returns strings."""
        a = self.tokens[index]
        if a.startswith('"') and a.endswith('"'):
            return a[1:-1]
        else:
            return a


class QschEditor(BaseSchematic):
    """Class made to update directly QSCH files. It is a subclass of BaseSchematic, so it can be used to
    update the netlist and the parameters of the simulation. It can also be used to update the components.

    :param qsch_file: Path to the QSCH file to be edited
    :type qsch_file: str
    :keyword create_blank: If True, the file will be created from scratch. If False, the file will be read and parsed
    """

    simulator_lib_paths: List[str] = Qspice.get_default_library_paths()    
    """ This is initialised with typical locations found for QSPICE.
    You can (and should, if you use wine), call `prepare_for_simulator()` once you've set the executable paths.
    
    :meta hide-value:
    """
        
    def __init__(self, qsch_file: str, create_blank: bool = False):
        super().__init__()
        self._qsch_file_path = Path(qsch_file)
        self.schematic = None
        # read the file into memory
        self.reset_netlist(create_blank)

    @property
    def circuit_file(self) -> Path:
        # docstring inherited from BaseSchematic
        return self._qsch_file_path

    def save_as(self, qsch_filename: Union[str, Path]) -> None:
        """
        Saves the schematic to a QSCH file. The file is saved in cp1252 encoding.
        """
        if self.updated or Path(qsch_filename) != self._qsch_file_path:
            with open(qsch_filename, 'w', encoding="cp1252") as qsch_file:
                _logger.info(f"Writing QSCH file {qsch_file}")
                for c in QSCH_HEADER:
                    qsch_file.write(chr(c))
                qsch_file.write(self.schematic.out(0))
                qsch_file.write('\n')  # Terminates the new line
            if Path(qsch_filename) == self._qsch_file_path:
                self.updated = False
        # now checks if there are subcircuits that need to be saved
        for component in self.components.values():
            if "_SUBCKT" in component.attributes:
                sub_circuit = component.attributes["_SUBCKT"]
                if sub_circuit.updated:
                    sub_circuit.save_as(sub_circuit._qsch_file_path)

    def write_spice_to_file(self, netlist_file: TextIO):
        """
        Appends the netlist to a file buffer.

        :param netlist_file: The file buffer to save the netlist
        :type netlist_file: TextIO
        :return: Nothing
        """
        libraries_to_include = []
        subcircuits_to_write = OrderedDict()

        for refdes, component in self.components.items():
            component: SchematicComponent
            item_tag = component.attributes['tag']
            disabled = not component.attributes['enabled']

            symbol_tags = item_tag.get_items('symbol')
            if len(symbol_tags) != 1 or disabled:
                continue
            symbol_tag = symbol_tags[0]
            if len(symbol_tag.tokens) > 1:
                symbol = symbol_tag.get_text_attr(1)
                typ = symbol_tag.get_text('type')
            else:
                symbol = 'X'
                typ = 'X'

            texts = symbol_tag.get_items('text')
            parameters = ""
            if len(texts) > 2:
                for text in texts[2:]:
                    parameters += " " + decap(text.get_text_attr(QSCH_TEXT_STR_ATTR))

            ports = component.ports.copy()
            if typ in ('¥', 'Ã'):
                if len(ports) < 16:
                    ports += ['¥'] * (16 - len(ports))

            nets = " ".join(ports)

            if typ == 'X':
                model = texts[1].get_text_attr(QSCH_TEXT_STR_ATTR)

                # schedule to write .SUBCKT clauses at the end
                if model not in subcircuits_to_write:
                    pins = symbol_tag.get_items("pin")
                    sub_ports = " ".join(pin.get_attr(QSCH_SYMBOL_PIN_NET) for pin in pins)
                    subcircuits_to_write[model] = (
                        component.attributes['_SUBCKT'],  # the subcircuit schematic is saved
                        sub_ports,  # and also storing the port position now, so to save time later.
                    )
                nets = " ".join(component.ports)
                netlist_file.write(f'{refdes} {nets} {model}{parameters}\n')

            elif typ in ('QP', 'QN'):
                model = texts[1].get_text_attr(QSCH_TEXT_STR_ATTR)
                if symbol == 'NPNS' or symbol == 'PNPS' or symbol == 'LPNP':
                    ports[3] = '[' + ports[3] + ']'
                    nets = ' '.join(ports)
                    hack = 'PNP' if 'PNP' in symbol else 'NPN'
                    netlist_file.write(f'{refdes} {nets} {model} {hack}{parameters}\n')
                else:
                    netlist_file.write(f'{refdes} {nets} [0] {model} {symbol}{parameters}\n')
            elif typ in ('MN', 'MP'):
                model = texts[1].get_text_attr(QSCH_TEXT_STR_ATTR)
                if symbol == 'NMOSB' or symbol == 'PMOSB':
                    symbol = symbol[0:4]
                if len(ports) == 3:
                    netlist_file.write(f'{refdes} {nets} {ports[2]} {model} {symbol}{parameters}\n')
                else:
                    netlist_file.write(f'{refdes} {nets} {model} {symbol}{parameters}\n')
            elif typ == 'T':
                model = decap(texts[1].get_text_attr(QSCH_TEXT_STR_ATTR))
                netlist_file.write(f'{refdes} {nets} {model}{parameters}\n')
            elif typ in ('JN', 'JP'):
                model = decap(texts[1].get_text_attr(QSCH_TEXT_STR_ATTR))
                if symbol.startswith('Pwr'):  # Hack alert. I don't know why the symbol is Pwr
                    symbol = symbol[3:]  # remove the Pwr from the symbol
                netlist_file.write(f'{refdes} {nets} {model} {symbol}{parameters}\n')
            elif typ == '×':
                model = decap(texts[1].get_text_attr(QSCH_TEXT_STR_ATTR))
                netlist_file.write(f'{refdes} «{nets}» {model}{parameters}\n')
            elif typ in ('ZP', 'ZN'):
                model = texts[1].get_text_attr(QSCH_TEXT_STR_ATTR)
                netlist_file.write(f'{refdes} {nets} {model} {symbol}{parameters}\n')
            else:
                value = texts[1].get_text_attr(QSCH_TEXT_STR_ATTR)
                netlist_file.write(f'{refdes} {nets} {value}{parameters}\n')
                # else:
                #     netlist_file.write(f'{symbol}†{refdes} {nets} {value}\n')

            library_tags = symbol_tag.get_items('library')
            for lib in library_tags:
                library_name = lib.get_text_attr(2)
                if library_name not in libraries_to_include:
                    libraries_to_include.append(library_name)

        for sub_circuit in subcircuits_to_write:
            sub_circuit_schematic, ports = subcircuits_to_write[sub_circuit]
            netlist_file.write("\n")
            netlist_file.write(f".subckt {sub_circuit} {ports}\n")
            sub_circuit_schematic.write_spice_to_file(netlist_file)
            netlist_file.write(f".ends {sub_circuit}\n")
            netlist_file.write("\n")

        for directive in self.directives:
            for line in directive.text.split('\\n'):
                if directive.type != TextTypeEnum.COMMENT:  # Comments are not written to the netlist
                    netlist_file.write(line.strip() + '\n')

        for library in libraries_to_include:
            library_path = self._qsch_file_find(library)
            if library_path is None:
                netlist_file.write(f'.lib {library}\n')
            else:
                if sys.platform.startswith("win"):
                    from spicelib.utils.windows_short_names import get_short_path_name
                    netlist_file.write(f'.lib {get_short_path_name(os.path.abspath(library_path))}\n')
                else:
                    netlist_file.write(f'.lib {os.path.abspath(library_path)}\n')

        # Note: the .END or .ENDCKT must be inserted by the calling function

    def save_netlist(self, run_netlist_file: Union[str, Path]) -> None:
        if isinstance(run_netlist_file, str):
            run_netlist_file = Path(run_netlist_file)

        if self.schematic is None:
            _logger.error("Empty Schematic information")
            return
        if run_netlist_file.suffix == '.qsch':
            self.save_as(run_netlist_file)
        elif run_netlist_file.suffix in ('.net', '.cir'):
            with open(run_netlist_file, 'w', encoding="cp1252") as netlist_file:
                _logger.info(f"Writing NET file {run_netlist_file}")
                netlist_file.write(f'* {os.path.abspath(self._qsch_file_path.as_posix())}\n')
                self.write_spice_to_file(netlist_file)
                netlist_file.write('.end\n')

    def _find_pin_position(self, comp_pos, orientation: int, pin: QschTag) -> Tuple[int, int]:
        """Returns the net name at the pin position"""
        pin_pos = pin.get_attr(1)
        hyp = (pin_pos[0] ** 2 + pin_pos[1] ** 2) ** 0.5
        if orientation % 2:
            # in 45º rotations the component is 1.414 times larger
            hyp *= 1.414
        if 0 <= orientation <= 7:
            theta = math.atan2(pin_pos[1], pin_pos[0]) + math.radians(orientation * 45)
            x = comp_pos[0] + round(hyp * math.cos(theta), -2)  # round to multiple of 100
            y = comp_pos[1] + round(hyp * math.sin(theta), -2)
        elif 8 <= orientation <= 15:
            # The component is mirrored on the X axis
            theta = math.atan2(pin_pos[1], pin_pos[0]) + math.radians((orientation - 8) * 45)
            x = comp_pos[0] - round(hyp * math.cos(theta), -2)  # round to multiple of 100
            y = comp_pos[1] + round(hyp * math.sin(theta), -2)
        else:
            raise ValueError(f"Invalid orientation: {orientation}")
        return x, y

    def _find_net_at_position(self, x, y) -> Optional[str]:
        """Returns the net name at the given position"""
        for net in self.schematic.get_items('net'):  # Connection to ports, grounds and nets
            if net.get_attr(1) == (x, y):
                net_name = net.get_attr(5)  # Found the net
                return '0' if net_name == 'GND' else net_name
        for wire in self.schematic.get_items('wire'):  # Connection to wires
            if wire.get_attr(1) == (x, y) or wire.get_attr(2) == (x, y):
                net_name = wire.get_attr(3)  # Found the net
                return '0' if net_name == 'GND' else net_name
        return None

    def reset_netlist(self, create_blank: bool = False) -> None:
        """
        If create_blank is True, it creates a blank netlist.

        If False, it reads the netlist from the file into memory. If the file does not exist, it raises a FileNotFoundError.

        All previous edits done to the netlist are lost.

        :param create_blank: If True, the file will be created from scratch. If False, the file will be read and parsed
        """
        super().reset_netlist(create_blank)
        if not create_blank:
            if not self._qsch_file_path.exists():
                raise FileNotFoundError(f"File {self._qsch_file_path} not found")
            with open(self._qsch_file_path, 'r', encoding="cp1252") as qsch_file:
                _logger.info(f"Reading QSCH file {self._qsch_file_path}")
                stream = qsch_file.read()
            self._parse_qsch_stream(stream)

    def _parse_qsch_stream(self, stream):
        """Parses the QSCH file stream"""
        self.components.clear()
        _logger.debug("Parsing QSCH file")
        header = tuple(ord(c) for c in stream[:4])

        if header != QSCH_HEADER:
            raise QschReadingError("Missing header. The QSCH file should start with: " +
                                   f"{' '.join(f'{c:02X}' for c in QSCH_HEADER)}")

        schematic, _ = QschTag.parse(stream, 4)
        self.schematic = schematic
        highest_net_number = 0
        behavior_pin_counter = 0
        unconnected_pins = {}  # Storing the components that have floating pins

        for net in self.schematic.get_items('net'):
            # process nets
            x, y = net.get_attr(QSCH_NET_POS)
            # TODO: Get the remaining attributes Rotation, size, color, etc...
            # rotation = net.get_attr(QSCH_NET_ROTATION)
            net_name = net.get_attr(QSCH_NET_STR_ATTR)
            self.labels.append(Text(Point(x, y), net_name, type=TextTypeEnum.LABEL))

        for wire in self.schematic.get_items('wire'):
            # process wires
            x1, y1 = wire.get_attr(QSCH_WIRE_POS1)
            x2, y2 = wire.get_attr(QSCH_WIRE_POS2)
            net = wire.get_attr(QSCH_WIRE_NET)
            # Check if the net is of the format N##, if it is get the net number
            if net.startswith('N'):
                try:
                    net_no = int(net[1:])
                    if net_no > highest_net_number:
                        highest_net_number = net_no
                except ValueError:
                    pass
            self.wires.append(Line(Point(x1, y1), Point(x2, y2), net=net))

        components = self.schematic.get_items('component')
        for component in components:
            symbol: QschTag = component.get_items('symbol')[0]
            texts = symbol.get_items('text')
            if len(texts) < 2:
                raise RuntimeError(f"Missing texts in component at coordinates {component.get_attr(1)}")
            refdes = texts[QSCH_SYMBOL_TEXT_REFDES].get_attr(QSCH_TEXT_STR_ATTR)
            value = texts[QSCH_SYMBOL_TEXT_VALUE].get_attr(QSCH_TEXT_STR_ATTR)
            sch_comp = SchematicComponent(self, refdes)
            sch_comp.reference = refdes
            x, y = position = component.get_attr(QSCH_COMPONENT_POS)
            orientation = component.get_attr(QSCH_COMPONENT_ROTATION)
            sch_comp.position = Point(x, y)
            sch_comp.rotation = orientation * 45
            sch_comp.attributes['type'] = symbol.get_text('type', "X")  # Assuming a sub-circuit
            sch_comp.attributes['description'] = symbol.get_text('description', "No Description")
            sch_comp.attributes['value'] = value
            sch_comp.attributes['tag'] = component
            sch_comp.attributes['enabled'] = component.get_attr(QSCH_COMPONENT_ENABLED) == 0
            sch_comp.ports = []
            pins = symbol.get_items('pin')

            for pin in pins:
                x, y = self._find_pin_position(position, orientation, pin)
                net = self._find_net_at_position(x, y)
                # The pins that have "¥" are behavioral pins, they are not connected to any net, they will be connected
                # to a net later.
                if refdes[0] in ('¥', 'Ã'):
                    if (len(pin.tokens) > QSCH_SYMBOL_PIN_NET_BEHAVIORAL and
                            pin.get_attr(QSCH_SYMBOL_PIN_NET_BEHAVIORAL) == '¥'):
                        net = '¥'
                if net is None:
                    hash_key = (x, y)
                    if hash_key in unconnected_pins:
                        net = unconnected_pins[hash_key]
                    else:
                        _logger.info(f"Unconnected pin at {x},{y} in component {refdes}:{pin}")
                        if refdes[0] in ('¥', 'Ã'):  # Behavioral pins are not connected
                            net = f'¥{behavior_pin_counter:d}'
                            behavior_pin_counter += 1
                        else:
                            highest_net_number += 1
                            net = f'N{highest_net_number:02d}'
                        unconnected_pins[hash_key] = net
                sch_comp.ports.append(net)

            self.components[refdes] = sch_comp
            if refdes.startswith('X'):
                sub_circuit_name = value + os.path.extsep + 'qsch'
                sub_circuit_schematic_file = self._qsch_file_find(sub_circuit_name)
                if sub_circuit_schematic_file:
                    sub_schematic = QschEditor(sub_circuit_schematic_file)
                    sch_comp.attributes['_SUBCKT'] = sub_schematic  # Store it for future use.

        for text_tag in self.schematic.get_items('text'):
            x, y = text_tag.get_attr(QSCH_TEXT_POS)
            point = Point(x, y)
            text = text_tag.get_attr(QSCH_TEXT_STR_ATTR)
            text_size = text_tag.get_attr(QSCH_TEXT_SIZE)
            if text_tag.get_attr(QSCH_TEXT_COMMENT) == 1:
                type_text = TextTypeEnum.COMMENT
            elif text.startswith(QSCH_TEXT_INSTR_QUALIFIER):
                type_text = TextTypeEnum.DIRECTIVE
                text = text.lstrip(QSCH_TEXT_INSTR_QUALIFIER)  # Eliminates the qualifer from the text.
            else:
                type_text = TextTypeEnum.NULL

            # angle = text_tag.get_attr(QSCH_TEXT_ROTATION)  # TODO: Implement text Rotation

            text_obj = Text(
                point,
                text,
                text_size,
                type_text,
                # textAlignment,
                # verticalAlignment,
                # angle=angle,
            )
            self.directives.append(text_obj)

        for line_tag in self.schematic.get_items('line'):
            x1, y1 = line_tag.get_attr(QSCH_LINE_POS1)
            x2, y2 = line_tag.get_attr(QSCH_LINE_POS2)
            width = line_tag.get_attr(QSCH_LINE_WIDTH)
            line_type = line_tag.get_attr(QSCH_LINE_TYPE)
            color = line_tag.get_attr(QSCH_LINE_COLOR)
            line = Line(Point(x1, y1), Point(x2, y2))
            line.style = LineStyle(width, line_type, color)
            self.lines.append(line)

    def _get_text_matching(self, command, search_expression: re.Pattern):
        command_upped = command.upper()
        text_tags = self.schematic.get_items('text')
        for tag in text_tags:
            line = tag.get_attr(QSCH_TEXT_STR_ATTR)
            line = line.lstrip(QSCH_TEXT_INSTR_QUALIFIER)
            if line.upper().startswith(command_upped):
                match = search_expression.search(line)
                if match:
                    return tag, match
        else:
            return None, None

    def _qsch_file_find(self, filename) -> Optional[str]:
        return search_file_in_containers(filename, 
                                         *self.custom_lib_paths,
                                         os.path.split(self._qsch_file_path)[0],  # The directory where the script is located
                                         *self.simulator_lib_paths)

    def get_subcircuit(self, reference: str) -> 'QschEditor':
        subcircuit = self.get_component(reference)
        if '_SUBCKT' in subcircuit.attributes:  # Optimization: if it was already stored, return it
            return subcircuit.attributes['_SUBCKT']
        raise AttributeError(f"An associated subcircuit was not found for {reference}")

    def get_parameter(self, param: str) -> str:
        # docstring inherited from BaseEditor
        param_regex = re.compile(PARAM_REGEX(param), re.IGNORECASE)
        tag, match = self._get_text_matching(".PARAM", param_regex)
        if match:
            return match.group('value')
        else:
            raise ParameterNotFoundError(f"Parameter {param} not found in QSCH file")

    def set_parameter(self, param: str, value: Union[str, int, float]) -> None:
        # docstring inherited from BaseEditor
        param_regex = re.compile(PARAM_REGEX(param), re.IGNORECASE)
        tag, match = self._get_text_matching(".PARAM", param_regex)
        if match:
            _logger.debug(f"Parameter {param} found in QSCH file, updating it")
            if isinstance(value, (int, float)):
                value_str = format_eng(value)
            else:
                value_str = value
            text: str = tag.get_attr(QSCH_TEXT_STR_ATTR)
            match = param_regex.search(text)  # repeating the search, so we update the correct start/stop parameter
            start, stop = match.span()
            text = text[:start] + "{}={}".format(param, value_str) + text[stop:]
            tag.set_attr(QSCH_TEXT_STR_ATTR, text)
            _logger.info(f"Parameter {param} updated to {value_str}")
            _logger.debug(f"Text at {tag.get_attr(QSCH_TEXT_POS)} Updated to {text}")
        else:
            # Was not found so we need to add it,
            _logger.debug(f"Parameter {param} not found in QSCH file, adding it")
            x, y = self._get_text_space()
            tag, _ = QschTag.parse(
                f'«text ({x},{y}) 1 0 0 0x1000000 -1 -1 "{QSCH_TEXT_INSTR_QUALIFIER}.param {param}={value}"»'
            )
            self.schematic.items.append(tag)
            _logger.info(f"Parameter {param} added with value {value}")
            _logger.debug(f"Text added to {tag.get_attr(QSCH_TEXT_POS)} Added: {tag.get_attr(QSCH_TEXT_STR_ATTR)}")
        self.updated = True

    def _get_component_symbol(self, reference: str) -> Tuple["BaseSchematic", str, QschTag]:
        sub_circuit, ref = self._get_parent(reference)
        if ref not in sub_circuit.components:
            _logger.error(f"Component {ref} not found")
            raise ComponentNotFoundError(f"Component {ref} not found in Schematic file")

        component = sub_circuit.components[ref]
        comp_tag: QschTag = component.attributes['tag']
        symbol: QschTag = comp_tag.get_items('symbol')[0]
        return sub_circuit, ref, symbol

    def set_component_value(self, reference: str, value: Union[str, int, float]) -> None:
        # docstring inherited from BaseEditor
        if isinstance(value, str):
            value_str = value
        else:
            value_str = format_eng(value)
        self.set_element_model(reference, value_str)

    def set_element_model(self, device: str, model: str) -> None:
        # docstring inherited from BaseEditor
        sub_circuit, ref, symbol = self._get_component_symbol(device)
        texts = symbol.get_items('text')
        assert texts[QSCH_SYMBOL_TEXT_REFDES].get_attr(QSCH_TEXT_STR_ATTR) == ref
        texts[QSCH_SYMBOL_TEXT_VALUE].set_attr(QSCH_TEXT_STR_ATTR, model)
        sub_circuit.components[ref].attributes['value'] = model
        _logger.info(f"Component {device} updated to {model}")
        sub_circuit.updated = True

    def get_component_value(self, element: str) -> str:
        # docstring inherited from BaseEditor
        component = self.get_component(element)
        if "value" not in component.attributes:
            _logger.error(f"Component {element} does not have a Value attribute")
            raise ComponentNotFoundError(f"Component {element} does not have a Value attribute")
        return component.attributes["value"]

    def get_component_parameters(self, element: str) -> dict:
        """
        Returns the parameters of the component in a dictionary. Since QSpice stores attributes by their order of
        appearance on the QSCH file, some parameters may not be found if they are not in the standard format.
        If a line contains a parameter definition that is on the standard format, it will be parsed and stored in the
        dictionary. The key of the dictionary is the line number where the parameter was found.

        :param element: The reference of the component
        :type element: str
        :return: A dictionary with the parameters of the component
        :rtype: dict
        """
        _, _, symbol = self._get_component_symbol(element)
        texts = symbol.get_items('text')
        parameters = {}
        param_regex = re.compile(PARAM_REGEX(r'\w+'), re.IGNORECASE)
        for i in range(2, len(texts)):
            text = texts[i].get_attr(QSCH_TEXT_STR_ATTR)
            matches = param_regex.finditer(text)
            for match in matches:
                parameters[match.group('name')] = match.group('value')
            else:
                parameters[i] = text

        return parameters

    def set_component_parameters(self, element: str, **kwargs) -> None:
        """
        Sets the parameters of the component. If key parameters that are integers, they represent the line number
        where the parameter was found. If the key is a string, it represents the parameter name. If the parameter name
        already exists, it will be replaced. If not found, it will be added as a new text line.
        """
        sub_circuit, ref, symbol = self._get_component_symbol(element)
        texts = symbol.get_items('text')

        for key, value in kwargs.items():
            if isinstance(key, int):
                if key < 2 or key > len(texts):
                    raise ValueError(f"Invalid line number {key} for component {element}")
                if key == len(texts):
                    x, y = 0, 0  # TODO: Find a way to get the position of the last text
                    tag, _ = QschTag.parse(
                        f'«text ({x},{y}) 1 7 0 0x1000000 -1 -1 "{value}"»'
                    )
                    symbol.items.append(tag)

                else:
                    texts[key].set_attr(QSCH_TEXT_STR_ATTR, value)
                sub_circuit.updated = True
            else:
                found = False
                search_expression = re.compile(PARAM_REGEX(key), re.IGNORECASE)
                for text in texts[QSCH_SYMBOL_TEXT_VALUE:]:
                    text_value = text.get_attr(QSCH_TEXT_STR_ATTR)
                    match = search_expression.search(text_value)
                    if match:
                        start, stop = match.span()
                        text_value = text_value[:start] + f"{key}={value}" + text_value[stop:]
                        text.set_attr(QSCH_TEXT_STR_ATTR, text_value)
                        sub_circuit.updated = True
                        found = True
                        break
                if not found:
                    x, y = 0, 0  # TODO: Find a way to get the position of the last text
                    new_tag, _ = QschTag.parse(
                        f'«text ({x},{y}) 0.5 0 0 0x1000000 -1 -1 "{key}={value}"»'
                    )
                    # Inserting the new tag just after the last text and the first pin
                    last_text = 0
                    for i, tag in enumerate(symbol.items):
                        if tag.tag == 'pin':
                            last_text = i
                            break
                        elif tag.tag == 'text':
                            last_text = i + 1  # The new text should be inserted after the last text
                    if 0 <= last_text < len(symbol.items):
                        symbol.items.insert(last_text, new_tag)
                    else:
                        symbol.items.append(new_tag)
                    sub_circuit.updated = True

    def get_component_position(self, reference: str) -> Tuple[Point, ERotation]:
        # docstring inherited from BaseSchematic
        component = self.get_component(reference)
        return component.position, component.rotation

    def set_component_position(self, reference: str,
                               position: Union[Point, tuple],
                               rotation: Union[ERotation, int],
                               mirror: bool = False,
                               ) -> None:
        # docstring inherited from BaseSchematic
        component = self.get_component(reference)
        comp_tag: QschTag = component.attributes['tag']
        if isinstance(position, tuple):
            position = (position[0], position[1])
        elif isinstance(position, Point):
            position = (position.X, position.Y)
        else:
            raise ValueError("Invalid position object")
        if isinstance(rotation, ERotation):
            rot = rotation.value / 45
        elif isinstance(rotation, int):
            rot = (rotation % 360) // 45
            if mirror:
                rot += 8
        else:
            raise ValueError("Invalid rotation parameter")

        comp_tag.set_attr(QSCH_COMPONENT_POS, position)
        comp_tag.set_attr(QSCH_COMPONENT_ROTATION, rot)
        component.position = position
        component.rotation = rotation

    def get_components(self, prefixes='*') -> list:
        # docstring inherited from BaseEditor
        if prefixes == '*':
            return list(self.components.keys())
        return [k for k in self.components.keys() if k[0] in prefixes]

    def remove_component(self, designator: str):
        # docstring inherited from BaseEditor
        component = self.get_component(designator)
        comp_tag: QschTag = component.attributes['tag']
        self.schematic.items.remove(comp_tag)

    def _get_text_space(self):
        """
        Returns the coordinate on the Schematic File canvas where a text can be appended.
        """
        first = True
        for tag in self.schematic.items:
            if tag.tag in ('component', 'net', 'text'):
                x1, y1 = tag.get_attr(1)
                x2, y2 = x1, y1  # todo: the whole component primitives
            elif tag.tag == 'wire':
                x1, y1 = tag.get_attr(1)
                x2, y2 = tag.get_attr(2)
            else:
                continue  # this avoids executing the code below when no coordinates are found
            if first:
                min_x = min(x1, x2)
                max_x = max(x1, x2)
                min_y = min(y1, y2)
                max_y = max(y1, y2)
                first = False
            else:
                min_x = min(min_x, x1, x2)
                max_x = max(max_x, x1, x2)
                min_y = min(min_y, y1, y2)
                max_y = max(max_y, y1, y2)

        if first:
            return 0, 0  # If no coordinates are found, we return the origin
        else:
            return min_x, min_y - 240  # Setting the text in the bottom left corner of the canvas

    def add_instruction(self, instruction: str) -> None:
        # docstring inherited from BaseEditor
        instruction = instruction.strip()  # Clean any end of line terminators
        command = instruction.split()[0].upper()

        if command in UNIQUE_SIMULATION_DOT_INSTRUCTIONS:
            # Before adding new instruction, if it is a unique instruction, we just replace it
            for text_tag in self.schematic.get_items('text'):
                text = text_tag.get_attr(QSCH_TEXT_STR_ATTR)
                text = text.lstrip(QSCH_TEXT_INSTR_QUALIFIER)
                command = text.split()[0].upper()
                if command in UNIQUE_SIMULATION_DOT_INSTRUCTIONS:
                    text_tag.set_attr(QSCH_TEXT_STR_ATTR, QSCH_TEXT_INSTR_QUALIFIER + instruction)
                    return  # Job done, can exit this method

        elif command.startswith('.PARAM'):
            raise RuntimeError('The .PARAM instruction should be added using the "set_parameter" method')
        # If we get here, then the instruction was not found, so we need to add it
        x, y = self._get_text_space()
        tag, _ = QschTag.parse(f'«text ({x},{y}) 1 0 0 0x1000000 -1 -1 "{QSCH_TEXT_INSTR_QUALIFIER}{instruction}"»')
        self.schematic.items.append(tag)

    def remove_instruction(self, instruction: str) -> None:
        # docstring inherited from BaseEditor
        for text_tag in self.schematic.get_items('text'):
            text = text_tag.get_attr(QSCH_TEXT_STR_ATTR)
            if instruction in text:
                self.schematic.items.remove(text_tag)
                _logger.info(f'Instruction "{instruction}" removed')
                return  # Job done, can exit this method

        msg = f'Instruction "{instruction}" not found'
        _logger.error(msg)

    def remove_Xinstruction(self, search_pattern: str) -> None:
        # docstring inherited from BaseEditor
        regex = re.compile(search_pattern, re.IGNORECASE)
        instr_removed = False
        for text_tag in self.schematic.get_items('text'):
            text = text_tag.get_attr(QSCH_TEXT_STR_ATTR)
            text = text.lstrip(QSCH_TEXT_INSTR_QUALIFIER)
            if regex.match(text):
                self.schematic.items.remove(text_tag)
                _logger.info(f'Instruction "{text}" removed')
                instr_removed = True
        if not instr_removed:
            msg = f'Instruction matching "{search_pattern}" not found'
            _logger.error(msg)

    def copy_from(self, editor: 'BaseSchematic') -> None:
        # docstring inherited from BaseSchematic
        super().copy_from(editor)
        # We need to copy the schematic information
        if isinstance(editor, QschEditor):
            from copy import deepcopy
            self.schematic = deepcopy(editor.schematic)
        else:
            # Need to create a new schematic from the netlist
            self.schematic = QschTag('schematic')
            for ref, comp in self.components.items():
                cmpx = comp.position.X
                cmpy = comp.position.Y
                rotation = int(comp.rotation) // 45
                comp_tag, _ = QschTag.parse(f'«component ({cmpx},{cmpy}) {rotation} 0»')
                if 'symbol' in comp.attributes:
                    comp_tag.items.append(comp.attributes['symbol'])
                self.schematic.items.append(comp_tag)

            for labels in self.labels:
                label_tag, _ = QschTag.parse('«net (0,0) 1 13 0 "0"»')
                label_tag.set_attr(QSCH_NET_STR_ATTR, labels.text)
                label_tag.set_attr(QSCH_NET_POS, (labels.coord.X, labels.coord.Y))
                self.schematic.items.append(label_tag)

            for wire in self.wires:
                wire_tag, _ = QschTag.parse('«wire (0,0) (0,0) "0"»')
                wire_tag.set_attr(QSCH_WIRE_POS1, (wire.V1.X, wire.V1.Y))
                wire_tag.set_attr(QSCH_WIRE_POS2, (wire.V2.X, wire.V2.Y))
                # wire_tag.set_attr(QSCH_WIRE_NET, wire.net)
                self.schematic.items.append(wire_tag)

            for line in self.lines:
                line_tag, _ = QschTag.parse('«line (2000,1300) (3150,-100) 0 2 0xff0000 -1 -1»')
                line_tag.set_attr(QSCH_LINE_POS1, (line.V1.X, line.V1.Y))
                line_tag.set_attr(QSCH_LINE_POS2, (line.V2.X, line.V2.Y))
                # TODO: Implement the style to width, type and color
                # line_width = 0  # Default
                # line_type = 0  # Default
                # color = 0xff0000  # Default : Blue Color
                # line_width, line_type, color = line.style.split(' ')
                # line_tag.set_attr(QSCH_LINE_WIDTH, line_width)
                # line_tag.set_attr(QSCH_LINE_TYPE, line_type)
                # line_tag.set_attr(QSCH_LINE_COLOR, color)
                self.schematic.items.append(line_tag)

            for shape in self.shapes:
                # TODO: Implement the line type and width conversion from LTSpice to QSpice.
                if shape.name == "RECTANGLE" or shape.name == "rect":
                    shape_tag, _ = QschTag.parse('«rect (1850,1550) (3650,-400) 0 0 2 0xff0000 0x1000000 -1 0 -1»')
                    shape_tag.set_attr(QSCH_RECT_POS1, (shape.points[0].X, shape.points[0].Y))
                    shape_tag.set_attr(QSCH_RECT_POS2, (shape.points[1].X, shape.points[1].Y))
                    # shape_tag.set_attr(QSCH_RECT_LINE_TYPE, shape.line_style.pattern)
                    # shape_tag.set_attr(QSCH_RECT_LINE_WIDTH, shape.line_style.width)
                    # shape_tag.set_attr(QSCH_RECT_LINE_COLOR, shape.line_style.color)
                elif shape.name == "CIRCLE" or shape.name == "ellipse":
                    shape_tag, _ = QschTag.parse('«ellipse (2100,1150) (2650,150) 0 0 2 0xff0000 0x1000000 -1 -1»')
                    shape_tag.set_attr(QSCH_ELLIPSE_POS1, (shape.points[0].X, shape.points[0].Y))
                    shape_tag.set_attr(QSCH_ELLIPSE_POS2, (shape.points[1].X, shape.points[1].Y))
                    # shape_tag.set_attr(QSCH_ELLIPSE_LINE_COLOR, shape.line_style.color)
                    # shape_tag.set_attr(QSCH_ELLIPSE_FILL_COLOR, shape.fill)
                elif shape.name == "ARC" or shape.name == "arc3p":
                    shape_tag, _ = QschTag.parse('«arc3p (2700,300) (2250,1200) (2500,800) 0 2 0xff0000 -1 -1»')
                    # TODO: Implement the ARC shape correctly.
                    # In LTSpice the First two points defines the bounding box of the arc and the following
                    # two points define the start and end of the arc.
                    # QSpice uses the first two points to define the start and end of the arc and the third point
                    # to define the curvature of the arc.
                    if shape.name == "ARC":
                        assert len(shape.points) == 4, "Invalid LTSpice shape"
                        center_x = int((shape.points[0].X + shape.points[1].X) / 2)
                        center_y = int((shape.points[0].Y + shape.points[1].Y) / 2)
                        start_angle = math.atan2(shape.points[2].Y - center_y, shape.points[2].X - center_x)
                        end_angle = math.atan2(shape.points[3].Y - center_y, shape.points[3].X - center_x)
                        ellipse_width = abs(shape.points[1].X - shape.points[0].X)
                        ellipse_height = abs(shape.points[1].Y - shape.points[0].Y)
                        start_point_x = int(center_x + ellipse_width / 2 * math.cos(start_angle))
                        start_point_y = int(center_y + ellipse_height / 2 * math.sin(start_angle))
                        end_point_x = int(center_x + ellipse_width / 2 * math.cos(end_angle))
                        end_point_y = int(center_y + ellipse_height / 2 * math.sin(end_angle))
                        shape_tag.set_attr(QSCH_ARC3P_POS1, (start_point_x, start_point_y))
                        shape_tag.set_attr(QSCH_ARC3P_POS2, (end_point_x, end_point_y))
                        shape_tag.set_attr(QSCH_ARC3P_POS3, (center_x, center_y))
                    else:
                        shape_tag.set_attr(QSCH_ARC3P_POS1, (shape.points[0].X, shape.points[0].Y))
                        shape_tag.set_attr(QSCH_ARC3P_POS2, (shape.points[1].X, shape.points[1].Y))
                        shape_tag.set_attr(QSCH_ARC3P_POS3, (shape.points[2].X, shape.points[2].Y))
                        # shape_tag.set_attr(QSCH_ARC3P_LINE_COLOR, shape.line_style.color)
                else:
                    print(f"Invalid shape {shape.name}. Being ignored. Ask Developper to implement this")

                self.schematic.items.append(shape_tag)

            for text in self.directives:
                text_tag, _ = QschTag.parse('«text (0,0) 1 7 0 0x1000000 -1 -1 "text"»')
                text_tag.set_attr(QSCH_TEXT_STR_ATTR, QSCH_TEXT_INSTR_QUALIFIER + text.text)
                text_tag.set_attr(QSCH_TEXT_POS, (text.coord.X, text.coord.Y))
                self.schematic.items.append(text_tag)



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        run_task.py
# Purpose:     Class used for a spice tool using a process call
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     23-12-2016
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
"""
Internal classes not to be used directly by the user
"""
__author__ = "Nuno Canto Brum <nuno.brum@gmail.com>"
__copyright__ = "Copyright 2023, Fribourg Switzerland"

from pathlib import Path
import sys
import threading
import time
import traceback
from time import sleep
from typing import Callable, Union, Any, Tuple, Type
import logging
_logger = logging.getLogger("spicelib.RunTask")

from .process_callback import ProcessCallback
from .simulator import Simulator

END_LINE_TERM = '\n'

if sys.version_info.major >= 3 and sys.version_info.minor >= 6:
    clock_function = time.time
else:
    clock_function = time.clock


def format_time_difference(time_diff):
    """Formats the time difference in a human-readable format, stripping the hours or minutes if they are zero"""
    seconds_difference = int(time_diff)
    milliseconds = int((time_diff - seconds_difference) * 1000)
    hours, remainder = divmod(seconds_difference, 3600)
    minutes, seconds = divmod(remainder, 60)

    if hours == 0:
        if minutes == 0:
            return f"{int(seconds):02d}.{milliseconds:04d} secs"
        else:
            return f"{int(minutes):02d}:{int(seconds):02d}.{milliseconds:04d}"
    else:
        return f"{int(hours):02d}:{int(minutes):02d}:{int(seconds):02d}.{milliseconds:04d}"


class RunTask(threading.Thread):
    """This is an internal Class and should not be used directly by the User."""

    def __init__(self, simulator: Type[Simulator], runno, netlist_file: Path,
                 callback: Union[Type[ProcessCallback], Callable[[Path, Path], Any]],
                 callback_args: dict = None,
                 switches: Any = None, timeout: float = None, verbose=False):

        super().__init__(name=f"RunTask#{runno}")
        self.start_time = None
        self.stop_time = None
        self.verbose = verbose
        self.switches = switches
        self.timeout = timeout  # Thanks to Daniel Phili for implementing this
        self.simulator = simulator
        self.runno = runno
        self.netlist_file = netlist_file
        self.callback = callback
        self.callback_args = callback_args
        self.retcode = -1  # Signals an error by default
        self.raw_file = None
        self.log_file = None
        self.callback_return = None

    def print_info(self, logger_fun, message):
        message = f"RunTask #{self.runno}:{message}"
        logger_fun(message)
        if self.verbose:
            print(f"{time.asctime()} {logger_fun.__name__}: {message}{END_LINE_TERM}")

    def run(self):
        # Running the Simulation

        self.start_time = clock_function()
        self.print_info(_logger.info, ": Starting simulation %d: %s" % (self.runno, self.netlist_file))

        # start execution
        self.retcode = self.simulator.run(self.netlist_file.absolute().as_posix(), self.switches, self.timeout)
        self.stop_time = clock_function()
        # print simulation time with format HH:MM:SS.mmmmmm

        # Calculate the time difference
        sim_time = format_time_difference(self.stop_time - self.start_time)
        # Format the time difference
        self.log_file = self.netlist_file.with_suffix('.log')

        # Cleanup everything
        if self.retcode == 0:
            self.raw_file = self.netlist_file.with_suffix(self.simulator.raw_extension)
            if self.raw_file.exists() and self.log_file.exists():
                # simulation successful
                self.print_info(_logger.info, "Simulation Successful. Time elapsed: %s" % sim_time)

                if self.callback:
                    if self.callback_args is not None:
                        callback_print = ', '.join([f"{key}={value}" for key, value in self.callback_args.items()])
                    else:
                        callback_print = ''
                    self.print_info(_logger.info, "Simulation Finished. Calling...{}(rawfile, logfile{})".format(
                        self.callback.__name__, callback_print))
                    try:
                        if self.callback_args is not None:
                            return_or_process = self.callback(self.raw_file, self.log_file, **self.callback_args)
                        else:
                            return_or_process = self.callback(self.raw_file, self.log_file)
                    except Exception:
                        error = traceback.format_exc()
                        self.print_info(_logger.error, error)
                    else:
                        if isinstance(return_or_process, ProcessCallback):
                            proc = return_or_process
                            proc.start()
                            self.callback_return = proc.queue.get()
                            proc.join()
                        else:
                            self.callback_return = return_or_process
                    finally:
                        callback_start_time = self.stop_time
                        self.stop_time = clock_function()
                        self.print_info(_logger.info, "Callback Finished. Time elapsed: %s" % format_time_difference(
                            self.stop_time - callback_start_time))
                else:
                    self.print_info(_logger.info, 'Simulation Finished. No Callback function given')
            else:
                self.print_info(_logger.error, "Simulation Raw file or Log file were not found")
        else:
            # simulation failed
            self.print_info(_logger.error, ": Simulation Aborted. Time elapsed: %s" % sim_time)
            if self.log_file.exists():
                self.log_file = self.log_file.replace(self.log_file.with_suffix('.fail'))

    def get_results(self) -> Union[None, Any, Tuple[str, str]]:
        """
        Returns the simulation outputs if the simulation and callback function has already finished.
        If the simulation is not finished, it simply returns None. If no callback function is defined, then
        it returns a tuple with (raw_file, log_file). If a callback function is defined, it returns whatever
        the callback function is returning.
        """
        if self.is_alive():
            return None

        if self.retcode == 0:  # All finished OK
            if self.callback:
                return self.callback_return
            else:
                return self.raw_file, self.log_file
        else:
            return '', ''

    def wait_results(self) -> Union[Any, Tuple[str, str]]:
        """
        Waits for the completion of the task and returns a tuple with the raw and log files.
        :returns: Tuple with the path to the raw file and the path to the log file
        :rtype: tuple(str, str)
        """
        while self.is_alive() or self.retcode == -1:
            sleep(0.1)
        return self.get_results()



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        process_callback.py
# Purpose:     Being able to execute callback in a separate process
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     23-04-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
"""

"""
from __future__ import annotations

from multiprocessing import Process, Queue
from typing import Any


class ProcessCallback(Process):
    """
    Wrapper for the callback function
    """
    def __init__(self, raw, log, group=None, name=None, *, daemon: bool | None = ...,
                 **kwargs) -> None:
        super().__init__(group=group, name=name, daemon=daemon)
        self.queue = Queue()
        self.raw_file = raw
        self.log_file = log
        self.kwargs = kwargs

    def run(self):
        ret = self.callback(self.raw_file, self.log_file, **self.kwargs)
        if ret is None:
            ret = "Callback doesn't return anything"
        self.queue.put(ret)

    @staticmethod
    def callback(raw_file, log_file, **kwargs) -> Any:
        """This function needs to be overriden"""
        ...



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        qspice_simulator.py
# Purpose:     Represents QSPICE
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     26-08-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
import sys
import os

from pathlib import Path
from typing import Union
import logging
from ..sim.simulator import Simulator, run_function, SpiceSimulatorError

_logger = logging.getLogger("spicelib.QSpiceSimulator")


class Qspice(Simulator):
    """Stores the simulator location and command line options and is responsible for generating netlists and running
    simulations."""
    
    raw_extension = '.qraw'
    """:meta private:"""
    
    #
    # In QSPICE all traces have double precision. This means that qraw files are not compatible
    # with LTSPICE
    
    # windows paths (that are also valid for wine)
    # Please note that os.path.expanduser and os.path.join are sensitive to the style of slash.
    # Placed in order of preference. The first to be found will be used.
    _spice_exe_win_paths = ["~/Qspice/QSPICE64.exe",
                            "~/AppData/Local/Programs/Qspice/QSPICE64.exe",
                            "C:/Program Files/QSPICE/QSPICE64.exe"]
    
    # the default lib paths, as used by get_default_library_paths
    _default_lib_paths = ["C:/Program Files/QSPICE",
                          "~/Documents/QSPICE"]

    """Searches on the any usual locations for a simulator"""
    # defaults:
    spice_exe = []
    process_name = None  
        
    if sys.platform == "linux" or sys.platform == "darwin":
        # status mid 2024: Qspice has limited support for running under linux+wine, and none for MacOS+wine
        # TODO: when the situation gets more mature, add support for wine. See LTspice for an example.
        spice_exe = []
        process_name = None
    else:  # Windows (well, also aix, wasi, emscripten,... where it will fail.)
        for exe in _spice_exe_win_paths:
            if exe.startswith("~"):
                # expand here, as I use _spice_exe_win_paths also for linux, and expanding earlier will fail
                exe = os.path.expanduser(exe)
            if os.path.exists(exe):
                spice_exe = [exe]
                break        

    # fall through        
    if len(spice_exe) == 0:
        spice_exe = []
        process_name = None
    else:
        process_name = Simulator.guess_process_name(spice_exe[0])        
        _logger.debug(f"Found Qspice installed in: '{spice_exe}' ")

    qspice_args = {
        'ASCII'     : ['-ASCII'],  # Use ASCII file format for the output data(.qraw) file.
        'binary'    : ['-binary'],  # Use binary file format for the output data(.qraw) file.
        'BSIM1'    : ['-BSIM1'],  # Use the charge-conserving BSIM1 charge model for MOS1, MOS2, and MOS3.
        'Meyer'    : ['-Meyer'],  # Use the Meyer Capacitance model for MOS1, MOS2, and MOS3.
        'o'         : ['-o', '<path>'],  # Specify the name of a file for the console output.
        # 'p'         : ['-p'],  # Take the netlist piped from stdin. Not used in this implementation.
        'ProtectSelections': ['-ProtectSelections', '<path>'],  # Protect sections marked with .prot/.unprot with encryption.
        'ProtectSubcircuits': ['-ProtectSubcircuits', '<path>'],  # Protect the body of subcircuits with encryption.
        'r'       : ['-r', '<path>'],  # Specify the name of the output data(.qraw) file.
    }
    """:meta private:"""

    @classmethod
    def valid_switch(cls, switch, path='') -> list:
        """
        Validates a command line switch. The following options are available for QSPICE:
        
            * 'ASCII': Use ASCII file format for the output data(.qraw) file.

            * 'binary': Use binary file format for the output data(.qraw) file.

            * 'BSIM1': Use the charge-conserving BSIM1 charge model for MOS1, MOS2, and MOS3.

            * 'Meyer': Use the Meyer Capacitance model for MOS1, MOS2, and MOS3.

            * 'o <path>': Specify the name of a file for the console output.

            * 'ProtectSelections <path>': Protect sections marked with .prot/.unprot with encryption.

            * 'ProtectSubcircuits <path>': Protect the body of subcircuits with encryption.

            * 'r <path>': Specify the name of the output data(.qraw) file.

        :param switch: switch to be added. If the switch is not on the list above, it should be correctly formatted with the preceding '-' switch
        :type switch: str
        :param path: path to the file related to the switch being given.
        :type path: str, optional
        :return: Nothing
        """
        if switch in cls.qspice_args:
            switches = cls.qspice_args[switch]
            switches = [switch.replace('<path>', path) for switch in switches]
            return switches
        else:
            raise ValueError("Invalid switch for class ")

    @classmethod
    def run(cls, netlist_file: Union[str, Path], cmd_line_switches: list = None, timeout: float = None, stdout=None, stderr=None) -> int:
        """Executes a Qspice simulation run.

        :param netlist_file: path to the netlist file
        :type netlist_file: Union[str, Path]
        :param cmd_line_switches: additional command line options. Best to have been validated by valid_switch(), defaults to None
        :type cmd_line_switches: list, optional
        :param timeout: If timeout is given, and the process takes too long, a TimeoutExpired exception will be raised, defaults to None
        :type timeout: float, optional
        :param stdout: control redirection of the command's stdout. Valid values are None, subprocess.PIPE, subprocess.DEVNULL, an existing file descriptor (a positive integer), and an existing file object with a valid file descriptor. With the default settings of None, no redirection will occur. 
        :type stdout: _FILE, optional
        :param stderr: Like stdout, but affecting the command's error output.
        :type stderr: _FILE, optional
        :raises SpiceSimulatorError: when the executable is not found.
        :raises NotImplementedError: when the requested execution is not possible on this platform.
        :return: return code from the process
        :rtype: int
        """
        if not cls.spice_exe:
            _logger.error("================== ALERT! ====================")
            _logger.error("Unable to find the QSPICE executable.")
            _logger.error("A specific location of the QSPICE can be set")
            _logger.error("using the create_from(<location>) class method")
            _logger.error("==============================================")
            raise SpiceSimulatorError("Simulator executable not found.")
        
        if cmd_line_switches is None:
            cmd_line_switches = []
        elif isinstance(cmd_line_switches, str):
            cmd_line_switches = [cmd_line_switches]
        netlist_file = Path(netlist_file)
                
        log_file = Path(netlist_file).with_suffix('.log').as_posix()
        cmd_run = cls.spice_exe + ['-o', log_file] + [netlist_file] + cmd_line_switches
        # start execution
        return run_function(cmd_run, timeout=timeout, stdout=stdout, stderr=stderr)



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        test_spice_editor.py
# Purpose:     Tool used validate the Spice Files Editor
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------

import os
import re
import sys
import unittest

sys.path.append(
    os.path.abspath((os.path.dirname(os.path.abspath(__file__)) + "/../")))  # add project root to lib search path

import spicelib
from spicelib.editor.base_editor import to_float
from spicelib.editor.spice_editor import component_replace_regexs

test_dir = '../examples/testfiles/' if os.path.abspath(os.curdir).endswith('unittests') else './examples/testfiles/'
golden_dir = './golden/' if os.path.abspath(os.curdir).endswith('unittests') else './unittests/golden/'
temp_dir = './temp/' if os.path.abspath(os.curdir).endswith('unittests') else './unittests/temp/'

if not os.path.exists(temp_dir):
    os.mkdir(temp_dir)


def check_value(test, regex, line, value, msg=None):
    r = regex.match(line)
    test.assertIsNotNone(r, "Accepted regex")
    value_str = r.group('value')
    if isinstance(value, str):
        value_test = value_str
    else:
        try:
            value_test = to_float(value_str)
        except ValueError:
            value_test = value_str
    if msg is None:
        msg = f"Pass {value} for {line}"

    if isinstance(value_test, float):
        test.assertAlmostEqual(value_test, value, 6, msg)
    else:
        test.assertEqual(value_test, value, msg)


class SpiceEditor_Test(unittest.TestCase):

    def setUp(self):
        self.edt = spicelib.editor.spice_editor.SpiceEditor(test_dir + "DC sweep.net")
        self.edt2 = spicelib.editor.spice_editor.SpiceEditor(test_dir + "opamptest.net")

    def test_component_editing_1(self):
        self.assertEqual(self.edt.get_component_value('R1'), '10k', "Tested R1 Value")  # add assertion here
        self.assertListEqual(self.edt.get_components(), ['Vin', 'R1', 'R2', 'D1'], "Tested get_components")  # add assertion here
        self.edt.set_component_value('R1', '33k')
        self.edt.save_netlist(temp_dir + 'test_components_output.net')
        self.equalFiles(temp_dir + 'test_components_output.net', golden_dir + 'test_components_output.net')
        self.assertEqual(self.edt.get_component_value('R1'), '33k', "Tested R1 Value")  # add assertion here
        self.edt.set_component_parameters('R1', Tc1=0, Tc2=0, pwr=None)
        self.edt.save_netlist(temp_dir + 'test_components_output_2.net')
        self.equalFiles(temp_dir + 'test_components_output_2.net', golden_dir + 'test_components_output_2.net')
        r1_params = self.edt.get_component_parameters('R1')
        for key, value in {'Tc1': 0, 'Tc2': 0}.items():
            self.assertEqual(r1_params[key], value, f"Tested R1 {key} Parameter")
        self.edt.remove_component('R1')
        self.edt.save_netlist(temp_dir + 'test_components_output_1.net')
        self.equalFiles(temp_dir + 'test_components_output_1.net', golden_dir + 'test_components_output_1.net')

    def test_component_editing_1_obj(self):
        self.assertListEqual(self.edt.get_components(), ['Vin', 'R1', 'R2', 'D1'], "Tested get_components")
        r1 = self.edt['R1']
        self.assertEqual(r1.value_str, '10k', "Tested R1 Value")
        self.assertEqual(r1.value, 10000, "Tested R1 Numeric Value")
        self.assertListEqual(r1.ports, ['in', 'out'], "Tested R1 Nodes")
        r1.value = '33k'
        self.assertEqual(r1.value_str, '33k', "Tested R1 Value")
        self.edt.save_netlist(temp_dir + 'test_components_output.net')
        self.equalFiles(temp_dir + 'test_components_output.net', golden_dir + 'test_components_output.net')
        self.assertEqual(self.edt['R1'].value_str, '33k', "Tested R1 Value")
        r1['Tc1'] = 0
        r1['Tc2'] = 0
        r1['pwr'] =None
        self.assertEqual(r1.params['Tc1'], 0, "Tested R1 Tc1 Parameter")
        self.assertEqual(r1.params['Tc2'], 0, "Tested R1 Tc2 Parameter")
        self.edt.save_netlist(temp_dir + 'test_components_output_2.net')
        self.equalFiles(temp_dir + 'test_components_output_2.net', golden_dir + 'test_components_output_2.net')
        r1_params = self.edt.get_component_parameters('R1')
        for key, value in {'Tc1': 0, 'Tc2': 0}.items():
            self.assertEqual(r1_params[key], value, f"Tested R1 {key} Parameter")
            self.assertEqual(r1[key], value, f"Tested R1 {key} Parameter")
        self.edt.remove_component('R1')
        self.edt.save_netlist(temp_dir + 'test_components_output_1.net')
        self.equalFiles(temp_dir + 'test_components_output_1.net', golden_dir + 'test_components_output_1.net')

    def test_component_editing_2(self):
        self.assertEqual(self.edt2.get_component_value('V1'), '15', "Tested V1 Value")
        self.assertEqual(self.edt2.get_component_value('V3'), 'PWL(1u 0 +2n 1 +1m 1 +2n 0 +1m 0 +2n -1 +1m -1 +2n 0) AC 1', "Tested V3 Value")  # complex value, with parameters
        self.assertEqual(self.edt2.get_component_value('XU1'), 'level2', "Tested U1 Value")  # has parameters
        self.assertEqual(self.edt2.get_component_parameters('XU1')['Rin'], '501Meg', "Tested U1 Rin Value") # last in the list
        self.assertEqual(self.edt2.get_component_value('XU2'), 'AD549', "Tested U2 Value")  # no parameters
        self.assertListEqual(self.edt2.get_components(), ['V1', 'V2', 'V3', 'XU1', 'XU2'], "Tested get_components")
        self.edt2.set_component_value('V3', 'PWL(2u 0 +1p 1 +1m 1)')
        self.edt2.set_component_parameters('V3', Rser=1)  # first in the list
        self.edt2.set_component_value('XU1', 'level3')
        self.edt2.set_component_parameters('XU1', GBW='1Meg')  # somewhere in the list
        self.edt2.save_netlist(temp_dir + 'opamptest_output_1.net')
        self.equalFiles(temp_dir + 'opamptest_output_1.net', golden_dir + 'opamptest_output_1.net')

    def test_parameter_edit(self):
        self.assertEqual(self.edt.get_parameter('TEMP'), '0', "Tested TEMP Parameter")  # add assertion here
        self.edt.set_parameter('TEMP', 25)
        self.assertEqual(self.edt.get_parameter('TEMP'), '25', "Tested TEMP Parameter")  # add assertion here
        self.edt.save_netlist(temp_dir + 'test_parameter_output.net')
        self.equalFiles(temp_dir + 'test_parameter_output.net', golden_dir + 'test_parameter_output.net')
        self.edt.set_parameter('TEMP', 0)  # reset to 0
        self.assertEqual(self.edt.get_parameter('TEMP'), '0', "Tested TEMP Parameter")  # add assertion here

    def test_instructions(self):
        self.edt.add_instruction('.ac dec 10 1 100k')
        self.edt.add_instruction('.save V(vout)')
        self.edt.add_instruction('.save I(R1)')
        self.edt.add_instruction('.save I(R2)')
        self.edt.add_instruction('.save I(D1)')
        self.edt.save_netlist(temp_dir + 'test_instructions_output.net')
        self.equalFiles(temp_dir + 'test_instructions_output.net', golden_dir + 'test_instructions_output.net')
        self.edt.remove_instruction('.save I(R1)')
        self.edt.save_netlist(temp_dir + 'test_instructions_output_1.net')
        self.equalFiles(temp_dir + 'test_instructions_output_1.net', golden_dir + 'test_instructions_output_1.net')
        self.edt.remove_Xinstruction(r"\.save\sI\(.*\)")  # removes all .save instructions for currents
        self.edt.save_netlist(temp_dir + 'test_instructions_output_2.net')
        self.equalFiles(temp_dir + 'test_instructions_output_2.net', golden_dir + 'test_instructions_output_2.net')

    def equalFiles(self, file1, file2):
        with open(file1, 'r') as f1:
            lines1 = f1.readlines()
        with open(file2, 'r') as f2:
            lines2 = f2.readlines()
        self.assertEqual(len(lines1), len(lines2), "Files have different number of lines")
        for i, lines in enumerate(zip(lines1, lines2)):
            self.assertEqual(lines[0], lines[1], "Line %d" % i)

    def test_resistors(self):
        """Validates the RegEx expressions on the Spice Editor file"""

        # Resistors
        regex_r = component_replace_regexs['R']
        self.assertIsNone(regex_r.match('X12 N1 N2 10k'), "Invalid prefix")

        check_value(self, regex_r, "Rq N1 N2 10k", '10k')
        check_value(self, regex_r, "Rq N1 N2 10R3", 10.3)
        check_value(self, regex_r, "Rq N1 N2 10k5", 10500)
        check_value(self, regex_r, "Rq N1 N2 10K6", 10600)
        check_value(self, regex_r, "Rq N1 N2 11Meg", 11E6)
        check_value(self, regex_r, "Rq N1 N2 10Meg5", 10.5e6)
        check_value(self, regex_r, "Rq N1 N2 {param1 + Param2}", "{param1 + Param2}")

    def test_capacitors(self):
        regex_c = component_replace_regexs['C']
        self.assertIsNone(regex_c.match('X12 N1 N2 10k'), "Invalid prefix")

        check_value(self, regex_c, "C10 N1 N2 10u", 10e-6)
        check_value(self, regex_c, "CX N1 N2 1U", 1e-6)
        check_value(self, regex_c, "Cq N1 N2 1UF", 1e-6)
        check_value(self, regex_c, "Cq N1 N2 120p", 12.0E-12)
        check_value(self, regex_c, "Cq N1 N2 12pF", 12.0E-12)
        check_value(self, regex_c, "Cq N1 N2 10.3E-9", 10.3E-9)
        check_value(self, regex_c, "Cq N1 N2 12e-12", 1e-12)
        check_value(self, regex_c, "Cq N1 N2 {param*2+3}", "{param*2+3}")

    def test_inductors(self):
        regex_l = component_replace_regexs['L']
        self.assertIsNone(regex_l.match('X12 N1 N2 10k'), "Invalid prefix")

        check_value(self, regex_l, "L1 N1 N2 10u", 10e-6)
        check_value(self, regex_l, "L2 N1 N2 10uH", 10e-6)
        check_value(self, regex_l, "L3 N1 N2 1U", 1e-6)
        check_value(self, regex_l, "L4 N1 N2 15mH", 0.015)
        check_value(self, regex_l, "LUN N1 N2 120nH", 120e-9)
        check_value(self, regex_l, "L55 N1 N2 12n", 12.0e-9)
        check_value(self, regex_l, "LA N1 N2 10.3E-9", 10.3e-9)
        check_value(self, regex_l, "LABC N1 N2 12e-12", 1e-12)
        check_value(self, regex_l, "LBCD N1 N2 {param*2+3}", "{param*2+3}")
        self.assertEqual(regex_l.match('LBCD N1 N2 {param*2+3}').group('nodes'), " N1 N2",  "Tested Inductor Value")

    def test_diodes(self):
        regex_d = component_replace_regexs['D']
        self.assertIsNone(regex_d.match('X12 N1 N2 10k'), "Invalid prefix")
        self.assertEqual('1N4148',regex_d.match('D1 N1 N2 1N4148').group('value'),  "Tested Diode Value")
        self.assertEqual('D', regex_d.match('D1 N1 N2 D').group('value'),  "Tested Diode Model")

    def test_bipolar(self):
        regex_q = component_replace_regexs['Q']
        self.assertIsNone(regex_q.match('X12 N1 N2 10k'), "Invalid prefix")
        self.assertEqual('2N3904', regex_q.match('Q1 N1 N2 N3 2N3904').group('value'), "Tested Transistor Value")
        self.assertEqual('Q', regex_q.match('Q1 N1 N2 N3 Q').group('value'), "Tested Transistor Model")

    def test_mosfets(self):
        regex_m = component_replace_regexs['M']
        self.assertIsNone(regex_m.match('X12 N1 N2 10k'), "Invalid prefix")
        self.assertEqual('IRF540', regex_m.match('M1 N1 N2 N3 N4 IRF540').group('value'), "Tested MOSFET Value")
        self.assertEqual('IRF540', regex_m.match('M1 N1 N2 N3 IRF540').group('value'), "Tested MOSFET Value")
        self.assertEqual('M', regex_m.match('M1 N1 N2 N3 N4 M').group('value'), "Tested MOSFET Model")

    def test_subcircuits(self):
        regex_x = component_replace_regexs['X']
        self.assertIsNone(regex_x.match('R1 N1 N2 10k'), "Invalid prefix")
        self.assertEqual('SUB1', regex_x.match('X12 N1 N2 N3 SUB1').group('value'), "Tested Subcircuit Value")
        self.assertEqual('SUB1', regex_x.match('X12 N1 N2 N3 N4 SUB1 x=123 y=4u').group('value'), "Tested Subcircuit Value")
        self.assertEqual(' x=123 y=4u', regex_x.match('X12 N1 N2 N3 N4 SUB1 x=123 y=4u').group('params'), "Tested Subcircuit Parameters")
        self.assertEqual(' N1 N2 N3 N4', regex_x.match('X12 N1 N2 N3 N4 SUB1 x=123 y=4u').group('nodes'), "Tested Subcircuit Ports")

    def test_independent_sources(self):
        regex_v = component_replace_regexs['V']
        self.assertIsNone(regex_v.match('R1 N1 N2 10k'), "Invalid prefix")
        self.assertEqual('2 AC 1', regex_v.match('V1 NC_08 NC_09 2 AC 1 Rser=3').group('value'), "Tested Voltage Source Value")
        self.assertEqual(' Rser=3', regex_v.match('V1 NC_08 NC_09 2 AC 1 Rser=3').group('params'), "Tested Voltage Source Value")
        self.assertEqual('PWL(1u 0 +2n 1 +1m 1 +2n 0 +1m 0 +2n -1 +1m -1 +2n 0)',
                         regex_v.match('V1 N1 N2 PWL(1u 0 +2n 1 +1m 1 +2n 0 +1m 0 +2n -1 +1m -1 +2n 0) Rser=3 Cpar=4').group('value'), "Tested Voltage Source Value")
        self.assertEqual(regex_v.match('V1 N001 0 5').group('value'), '5', "Tested Voltage Source Value")
        self.assertEqual(regex_v.match('V1 N1 N2 5').group('value'), '5', "Tested Voltage Source Value")
        self.assertEqual(regex_v.match('V1 N1 N2 5V').group('value'), '5V', "Tested Voltage Source Value")
        self.assertEqual(regex_v.match('V1 N1 N2 {param}').group('value'), '{param}', "Tested Voltage Source Value")

        regex_i = component_replace_regexs['I']
        self.assertEqual('2 AC 1', regex_i.match('I1 NC_08 NC_09 2 AC 1 Rser=3').group('value'), "Tested Independent Current Source Value")
        self.assertEqual(' Rser=3', regex_i.match('I1 NC_08 NC_09 2 AC 1 Rser=3').group('params'), "Tested Independent Current Source Value")
        self.assertEqual('PWL(1u 0 +2n 1 +1m 1 +2n 0 +1m 0 +2n -1 +1m -1 +2n 0)',
                         regex_i.match('I1 N1 N2 PWL(1u 0 +2n 1 +1m 1 +2n 0 +1m 0 +2n -1 +1m -1 +2n 0) Rser=3 Cpar=4').group('value'), "Tested Voltage Source Value")
        self.assertEqual(regex_i.match('I1 N001 0 5').group('value'), '5', "Tested Independent Current Source Value")
        self.assertEqual(regex_i.match('I1 N1 N2 5').group('value'), '5', "Tested Independent Current Source Value")
        self.assertEqual(regex_i.match('I1 N1 N2 5V').group('value'), '5V', "Tested Independent Current Source Value")
        self.assertEqual(regex_i.match('I1 N1 N2 {param}').group('value'), '{param}', "Tested Independent Current Source Value")

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        raw_convert.py
# Purpose:     command line read of the raw file and output to CSV
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     17-01-2017
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
"""
Helper script to read the raw file and output to a CSV, Excel or Clipboard

Usage:
    raw_convert.py [options] <rawfile> <trace_list>

Options:
    -h --help               Show this screen
    -v --version            Show version
    -o --output=<file>      Output file name. Valid extensions are .csv, .xlsx
    -c --clipboard          Output to clipboard
    -s --separator=<sep>    Separator for CSV output [default: \t]

The script will read the raw file and output the specified traces to the chosen format.
In case of .TRAN simulations, the time column is always output. In case of .AC simulations
the frequency column is always output. The traces are specified in the <trace_list> as a
space separated list of trace names.
The V() qualifier is optional. If the trace name is not found, the script will try to
find the trace name with the V() qualifier. If the trace name is not found, the script
will try to find the trace name with the I() qualifier.

"""

from optparse import OptionParser
import clipboard
from spicelib.raw.raw_read import RawRead


def main():
    usage = "usage: %prog [options] <rawfile> <trace_list>"
    parser = OptionParser(usage=usage, version="%prog 0.1")
    parser.add_option("-o", "--output", dest="output", default=None,
                      help="Output file name.\n"
                           "Use .csv for CSV output, .xlsx for Excel output",
                      metavar="FILE")
    parser.add_option("-c", "--clipboard", dest="clipboard", action="store_true",
                      help="Output to clipboard", default=False)
    parser.add_option("-v", "--verbose", dest="verbose", action="store_true",
                      help="Verbose output", default=False)
    parser.add_option('-s', '--sep', dest='separator', default='\t',
                      help='Value separator for CSV output. Default: "\\t" <TAB>\n'
                           'Example: -d ";"'
                      )

    (options, args) = parser.parse_args()

    if len(args) < 1:
        print("Error: Missing arguments")
        parser.print_help()
        traces = '*'  # This is here just to avoid a warning on the IDE
        exit(1)
    elif len(args) < 2:
        traces = '*'
    else:
        traces = args[1:]

    rawfile = args[0]

    # Read the raw file
    if traces != '*':
        raw_data = RawRead(rawfile, '*', header_only=True, verbose=False)
        raw_traces = raw_data.get_trace_names()
        found_traces = []
        for trace in traces:
            if trace in raw_traces:
                found_traces.append(trace)
            else:
                if options.verbose:
                    print("Trace " + trace + " not found. Searching for V(" + trace + ")")
                if 'V(' + trace + ')' in raw_traces:
                    found_traces.append('V(' + trace + ')')
                else:
                    if options.verbose:
                        print("Trace V(" + trace + ") not found. Searching for I(" + trace + ")")
                    if 'I(' + trace + ')' in raw_traces:
                        found_traces.append('I(' + trace + ')')
                    else:
                        print("Warning: Trace " + trace + " not found")

        if len(found_traces) == 0:
            print("Error: No traces found")
            print("Available Traces:\n ")
            for trace in raw_traces:
                print("\t<" + trace + ">")
            exit(1)
        print("Reading traces: ", found_traces)
        raw_data = RawRead(rawfile, found_traces, verbose=options.verbose)
    else:
        raw_data = RawRead(rawfile, traces, verbose=options.verbose)

    # Output the file
    if options.output is None:
        data = raw_data.export()

        text = options.separator.join(data.keys()) + '\n'
        data_size = len(data[data.__iter__().__next__()])
        for i in range(data_size):
            text += options.separator.join([str(data[col][i]) for col in data.keys()]) + '\n'
        if options.clipboard:
            print(f"Copying to clipboard text with {len(text)} bytes")
            clipboard.copy(text)

        else:
            print(text)
    else:
        if options.output.endswith(".csv"):
            print("Writing CSV file...", end="")
            raw_data.to_csv(options.output, separator=options.separator, index=False)
            print("Done")
        elif options.output.endswith(".xlsx"):
            print("Writing Excel file...", end="")
            raw_data.to_excel(options.output, index=False)
            print("Done")
        else:
            print("Error: Unknown output format. Valid formats are '.csv' and '.xlsx'")
            parser.print_help()
            exit(1)
    exit(0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        run_server.py
# Purpose:     A Command Line Interface to run the LTSpice Server
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     10-08-2023
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
import sys

import time
import keyboard
import logging

from spicelib.client_server.sim_server import SimServer
from spicelib.simulators.ltspice_simulator import LTspice
simulator = LTspice

_logger = logging.getLogger("spicelib.SimServer")
_logger.setLevel(logging.DEBUG)
_logger.addHandler(logging.StreamHandler(sys.stdout))

print("Starting Server")
server = SimServer(simulator, parallel_sims=4, output_folder='./temp_server', port=9000)
print("Server Started. Press and hold 'q' to stop")

while server.running():
    time.sleep(0.2)
    # Check whether a key was pressed
    if keyboard.is_pressed('q'):
        server.stop_server()
        break



#!/usr/bin/env python
# coding=utf-8

# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        ltspice_simulator.py
# Purpose:     Represents a LTspice tool and it's command line options
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     23-12-2016
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
import sys
import os

from pathlib import Path
from typing import Union
import logging
from ..sim.simulator import Simulator, run_function, SpiceSimulatorError

_logger = logging.getLogger("spicelib.LTSpiceSimulator")


class LTspice(Simulator):
    """Stores the simulator location and command line options and is responsible for generating netlists and running
    simulations."""

    """Searches on the any usual locations for a simulator"""
    # windows paths (that are also valid for wine)
    # Please note that os.path.expanduser and os.path.join are sensitive to the style of slash.
    # Placed in order of preference. The first to be found will be used.
    _spice_exe_win_paths = ["~/AppData/Local/Programs/ADI/LTspice/LTspice.exe",
                            "~/Local Settings/Application Data/Programs/ADI/LTspice/LTspice.exe",
                            "C:/Program Files/ADI/LTspice/LTspice.exe",
                            "C:/Program Files/LTC/LTspiceXVII/XVIIx64.exe",
                            "C:/Program Files (x86)/LTC/LTspiceIV/scad3.exe"
                            ]
    
    # the default lib paths, as used by get_default_library_paths
    _default_lib_paths = ["~/AppData/Local/LTspice/lib",
                          "~/Documents/LtspiceXVII/lib/",
                          "~/Local Settings/Application Data/LTspice/lib"]
    
    # defaults:
    spice_exe = []
    process_name = None       
    
    if sys.platform == "linux" or sys.platform == "darwin":
        # Linux: look for wine and ltspice under wine.
        # MacOS: give preference to wine. If not found: look for native LTspice
        
        # Anything specified in environment variables?
        spice_folder = os.environ.get("LTSPICEFOLDER")
        spice_executable = os.environ.get("LTSPICEEXECUTABLE")

        if spice_folder and spice_executable:
            spice_exe = ["wine", os.path.join(spice_folder, spice_executable)]
            process_name = spice_executable
        elif spice_folder:
            spice_exe = ["wine", os.path.join(spice_folder, "/XVIIx64.exe")]
            process_name = "XVIIx64.exe"
        elif spice_executable:
            default_folder = os.path.expanduser(
                "~/.wine/drive_c/Program Files/LTC/LTspiceXVII"
            )
            spice_exe = ["wine", os.path.join(default_folder, spice_executable)]
            process_name = spice_executable
        else:
            # This is still "linux or darwin"
            # no environment variables was given. Do a search.
            for exe in _spice_exe_win_paths:
                # make the file path wine compatible
                # Note that wine also accepts paths like 'C:\users\myuser\...'.
                # BUT, if I do that, I would not be able to check for the presence of the exe.
                # So: expand everything.                 
                # Linux would use this: 
                #    '/home/myuser/.wine/drive_c/users/myuser/AppData/...'  for _spice_exe_win_paths[0]
                # or '/home/myuser/.wine/drive_c/Program Files/...'         for _spice_exe_win_paths[2]
                # MacOS would use this: 
                #    '/Users/myuser/.wine/drive_c/users/myuser/AppData/...' for _spice_exe_win_paths[0]    
                # or '/Users/myuser/.wine/drive_c/Program Files/...'        for _spice_exe_win_paths[2]  
                # Note that in the user path versions (_spice_exe_win_paths[0] and [1]), I have 2 expansions of the user name.
                if exe.startswith("~"):
                    exe = "C:/users/" + os.path.expandvars("${USER}" + exe[1:])
                # Now I have a "windows" path (but with forward slashes). Make it into a path under wine.
                exe = os.path.expanduser(exe.replace("C:/", "~/.wine/drive_c/"))
                
                if os.path.exists(exe):
                    spice_exe = ["wine", exe]
                    # Note that one easy method of killing a wine process is to run "wineserver -k", 
                    # but we kill via psutil....kill(), as that would also fit non-wine executions.
                    
                    break
            else:
                # The else block will not be executed if the loop is stopped by a break statement.
                # in case of MacOS, try the native LTspice as last resort
                if sys.platform == "darwin":
                    exe = '/Applications/LTspice.app/Contents/MacOS/LTspice'
                    if os.path.exists(exe):
                        spice_exe = [exe]
                                
    else:  # Windows (well, also aix, wasi, emscripten,... where it will fail.)
        for exe in _spice_exe_win_paths:
            if exe.startswith("~"):
                # expand here, as I use _spice_exe_win_paths also for linux, and expanding earlier will fail
                exe = os.path.expanduser(exe)
            if os.path.exists(exe):
                spice_exe = [exe]
                break
    
    # The following variables are not needed anymore. This also makes sphinx not mention them in the documentation.
    del exe
    if sys.platform == "linux" or sys.platform == "darwin":
        del spice_folder
        del spice_executable
            
    # fall through        
    if len(spice_exe) == 0:
        spice_exe = []
        process_name = None
    else:
        process_name = Simulator.guess_process_name(spice_exe[0])
        _logger.debug(f"Found LTspice installed in: '{spice_exe}' ")

    ltspice_args = {
        'alt'                : ['-alt'],  # Set solver to Alternate.
        'ascii'              : ['-ascii'],  # Use ASCII.raw files. Seriously degrades program performance.
        # 'batch'            : ['-b <path>'], # Used by run command: Run in batch mode.E.g. "ltspice.exe-b deck.cir" will leave the data infile deck.raw
        'big'                : ['-big'],  # Start as a maximized window.
        'encrypt'            : ['-encrypt'],
        # Encrypt a model library.For 3rd parties wishing to allow people to use libraries without
        # revealing implementation details. Not used by AnalogDevices models.
        'fastaccess'         : ['-FastAccess'],  # Batch conversion of a binary.rawfile to Fast Access format.
        'FixUpSchematicFonts': ['-FixUpSchematicFonts'],
        # Convert the font size field of very old user - authored schematic text to the modern default.
        'FixUpSymbolFonts'   : ['-FixUpSymbolFonts'],
        # Convert the font size field of very old user - authored symbols to the modern default.
        # See Changelog.txt for application hints.
        'ini'                : ['- ini', '<path>'],  # Specify an .ini file to use other than %APPDATA%\LTspice.ini
        'I'                  : ['-I<path>'],  # Specify a path to insert in the symbol and file search paths.
        # Must be the last specified option.
        # No space between "-I" and < path > is allowed.
        'max'                : ['-max'],  # Synonym for -big
        'netlist'            : ['-netlist'],  # Batch conversion of a schematic to a netlist.
        'norm'               : ['-norm'],  # Set solver to Normal.
        'PCBnetlist': ['-PCBnetlist'],  # Batch conversion of a schematic to a PCB format netlist.
        # 'run'              : ['-Run', '-b', '{path}'],  # Start simulating the schematic opened on the command line without
        # pressing the Run button.
        'SOI'                : ['-SOI'],  # Allow MOSFET's to have up to 7 nodes even in subcircuit expansion.
        'sync'               : ['-sync'],  # Update component libraries
        'uninstall'          : ['-uninstall'],  # Please don't. Executes one step of the uninstallation process.
    }
    """:meta private:"""

    @classmethod
    def using_macos_native_sim(cls) -> bool:
        """Tells if the simulator used is the MacOS native LTspice

        :return: True if the MacOS native LTspice is used, False otherwise (will also return False on Windows or Linux)
        :rtype: bool
        """
        return sys.platform == "darwin" and cls.spice_exe and "wine" not in cls.spice_exe[0].lower()

    @classmethod
    def valid_switch(cls, switch, path='') -> list:
        """
        Validates a command line switch. The following options are available for Windows/wine LTspice:

            * 'alt' : Set solver to Alternate.

            * 'ascii'     : Use ASCII.raw files. Seriously degrades program performance.

            * 'encrypt'   : Encrypt a model library.For 3rd parties wishing to allow people to use libraries without
                            revealing implementation details. Not used by AnalogDevices models.

            * 'fastaccess': Batch conversion of a binary.rawfile to Fast Access format.

            * 'FixUpSchematicFonts' : Convert the font size field of very old user - authored schematic text to the
                                    modern default.

            * 'FixUpSymbolFonts' : Convert the font size field of very old user - authored symbols to the modern
                default. See Changelog.txt for application hints.

            * 'ini <path>' : Specify an .ini file to use other than %APPDATA%\\LTspice.ini

            * 'I<path>' : Specify a path to insert in the symbol and file search paths. Must be the last specified
                option.

            * 'netlist'   :  Batch conversion of a schematic to a netlist.

            * 'normal'    :  Set solver to Normal.

            * 'PCBnetlist':  Batch conversion of a schematic to a PCB format netlist.

            * 'SOI'       :  Allow MOSFET's to have up to 7 nodes even in subcircuit expansion.

            * 'sync'      : Update component libraries

            * 'uninstall' :  Executes one step of the uninstallation process. Please don't.
            
            MacOS native LTspice accepts no command line switches (yet). 


        :param switch: switch to be added. If the switch is not on the list above, it should be correctly formatted with
                       the preceding '-' switch
        :type switch: str
        :param path: path to the file related to the switch being given.
        :type path: str, optional
        :return: Nothing
        """
        
        # See if the MacOS simulator is used. If so, check if I use the native simulator
        if cls.using_macos_native_sim():
            # this is the native LTspice. It has no useful command line switches (except '-b').
            raise ValueError("MacOS native LTspice does not support command line switches. Use it under wine for full support.")
            
        if switch in cls.ltspice_args:
            switches = cls.ltspice_args[switch]
            switches = [switch.replace('<path>', path) for switch in switches]
            return switches
        else:
            raise ValueError("Invalid switch for class ")

    @classmethod
    def run(cls, netlist_file: Union[str, Path], cmd_line_switches: list = None, timeout: float = None, stdout=None, stderr=None) -> int:
        """Executes a LTspice simulation run.

        :param netlist_file: path to the netlist file
        :type netlist_file: Union[str, Path]
        :param cmd_line_switches: additional command line options. Best to have been validated by valid_switch(), defaults to None
        :type cmd_line_switches: list, optional
        :param timeout: If timeout is given, and the process takes too long, a TimeoutExpired exception will be raised, defaults to None
        :type timeout: float, optional
        :param stdout: control redirection of the command's stdout. Valid values are None, subprocess.PIPE, subprocess.DEVNULL, an existing file descriptor (a positive integer), and an existing file object with a valid file descriptor. With the default settings of None, no redirection will occur. 
        :type stdout: _FILE, optional
        :param stderr: Like stdout, but affecting the command's error output.
        :type stderr: _FILE, optional
        :raises SpiceSimulatorError: when the executable is not found.
        :raises NotImplementedError: when the requested execution is not possible on this platform.
        :return: return code from the process
        :rtype: int
        """
        if not cls.spice_exe:
            _logger.error("================== ALERT! ====================")
            _logger.error("Unable to find a LTspice executable.")
            _logger.error("A specific location of the LTSPICE can be set")
            _logger.error("using the create_from(<location>) class method")
            _logger.error("==============================================")
            raise SpiceSimulatorError("Simulator executable not found.")
        
        if cmd_line_switches is None:
            cmd_line_switches = []
        elif isinstance(cmd_line_switches, str):
            cmd_line_switches = [cmd_line_switches]
        netlist_file = Path(netlist_file)        
        
        if sys.platform == "linux" or sys.platform == "darwin":
            if cls.using_macos_native_sim():
                # native MacOS simulator, which has its limitations
                if netlist_file.lower().endswith(".asc"):
                    raise NotImplementedError("MacOS native LTspice cannot run simulations on '.asc' files. Simulate '.net' or '.cir' files or use LTspice under wine.")
                
                cmd_run = cls.spice_exe + ['-b'] + [netlist_file.as_posix()] + cmd_line_switches
            else:
                # wine
                # Drive letter 'Z' is the link from wine to the host platform's root directory. 
                # Z: is needed for netlists with absolute paths, but will also work with relative paths.
                cmd_run = cls.spice_exe + ['-Run'] + ['-b'] + ['Z:' + netlist_file.as_posix()] + cmd_line_switches
        else:
            # Windows (well, also aix, wasi, emscripten,... where it will fail.)
            cmd_run = cls.spice_exe + ['-Run'] + ['-b'] + [netlist_file.as_posix()] + cmd_line_switches
        # start execution
        return run_function(cmd_run, timeout=timeout, stdout=stdout, stderr=stderr)

    @classmethod
    def create_netlist(cls, circuit_file: Union[str, Path], cmd_line_switches: list = None, stdout=None, stderr=None) -> Path:
        """Create a netlist out of the circuit file

        :param circuit_file: path to the circuit file
        :type circuit_file: Union[str, Path]
        :param cmd_line_switches: additional command line options. Best to have been validated by valid_switch(), defaults to None
        :type cmd_line_switches: list, optional
        :param stdout: control redirection of the command's stdout. Valid values are None, subprocess.PIPE, subprocess.DEVNULL, an existing file descriptor (a positive integer), and an existing file object with a valid file descriptor. With the default settings of None, no redirection will occur. 
        :type stdout: _FILE, optional
        :param stderr: Like stdout, but affecting the command's error output.
        :type stderr: _FILE, optional
        :raises NotImplementedError: when the requested execution is not possible on this platform.
        :raises RuntimeError: when the netlist cannot be created
        :return: path to the netlist produced
        :rtype: Path
        """
        # prepare instructions, two stages used to enable edits on the netlist w/o open GUI
        # see: https://www.mikrocontroller.net/topic/480647?goto=5965300#5965300
        if cmd_line_switches is None:
            cmd_line_switches = []
        elif isinstance(cmd_line_switches, str):
            cmd_line_switches = [cmd_line_switches]
        circuit_file = Path(circuit_file)
        
        if cls.using_macos_native_sim():
            # native MacOS simulator
            raise NotImplementedError("MacOS native LTspice does not have netlist generation capabilities. Use LTspice under wine.")
        
        cmd_netlist = cls.spice_exe + ['-netlist'] + [circuit_file.as_posix()] + cmd_line_switches
        error = run_function(cmd_netlist, stdout=stdout, stderr=stderr)

        if error == 0:
            netlist = circuit_file.with_suffix('.net')
            if netlist.exists():
                _logger.debug("OK")
                return netlist
        msg = "Failed to create netlist"
        _logger.error(msg)
        raise RuntimeError(msg)



#!/usr/bin/env python
# coding=utf-8

from spicelib.log.ltsteps import LTSpiceLogReader

data = LTSpiceLogReader("./testfiles/Batch_Test_AD820_15.log")

print("Number of steps  :", data.step_count)
step_names = data.get_step_vars()
meas_names = data.get_measure_names()

# Printing Headers
print(' '.join([f"{step:15s}" for step in step_names]), end='')  # Print steps names with no new line
print(' '.join([f"{name:15s}" for name in meas_names]), end='\n')
# Printing data
for i in range(data.step_count):
    print(' '.join([f"{data[step][i]:15}" for step in step_names]), end='')  # Print steps names with no new line
    print(' '.join([f"{data[name][i]:15}" for name in meas_names]), end='\n')  # Print Header

print("Total number of measures found :", data.measure_count)



#!/usr/bin/env python
# coding=utf-8
# -------------------------------------------------------------------------------
#
#  ███████╗██████╗ ██╗ ██████╗███████╗██╗     ██╗██████╗
#  ██╔════╝██╔══██╗██║██╔════╝██╔════╝██║     ██║██╔══██╗
#  ███████╗██████╔╝██║██║     █████╗  ██║     ██║██████╔╝
#  ╚════██║██╔═══╝ ██║██║     ██╔══╝  ██║     ██║██╔══██╗
#  ███████║██║     ██║╚██████╗███████╗███████╗██║██████╔╝
#  ╚══════╝╚═╝     ╚═╝ ╚═════╝╚══════╝╚══════╝╚═╝╚═════╝
#
# Name:        asc_to_qsch.py
# Purpose:     Convert an ASC file to a QSCH schematic
#
# Author:      Nuno Brum (nuno.brum@gmail.com)
#
# Created:     16-02-2024
# Licence:     refer to the LICENSE file
# -------------------------------------------------------------------------------
import os
import logging
import xml.etree.ElementTree as ET

from spicelib.editor.asy_reader import AsyReader
from spicelib.editor.asc_editor import AscEditor
from spicelib.editor.qsch_editor import QschEditor
from spicelib.utils.file_search import find_file_in_directory

_logger = logging.getLogger()


def main():
    """Converts an ASC file to a QSCH schematic"""
    import os.path
    from optparse import OptionParser

    opts = OptionParser(
        usage="usage: %prog [options] ASC_FILE [QSCH_FILE]",
        version="%prog 0.1")

    opts.add_option('-a', "--add", action="append", type="string", dest="path",
                    help="Add a path for searching for symbols")

    (options, args) = opts.parse_args()

    if len(args) < 1:
        opts.print_help()
        exit(-1)

    asc_file = args[0]
    if len(args) > 1:
        qsch_file = args[1]
    else:
        qsch_file = os.path.splitext(asc_file)[0] + ".qsch"

    search_paths = [] if options.path is None else options.path

    print(f"Using {qsch_file} as output file")
    convert_asc_to_qsch(asc_file, qsch_file, search_paths)


def convert_asc_to_qsch(asc_file, qsch_file, search_paths=[]):
    """Converts an ASC file to a QSCH schematic"""
    symbol_stock = {}
    # Open the ASC file
    asc_editor = AscEditor(asc_file)

    # import the conversion data from xml file
    # need first to find the file. It is in the same directory as the script
    parent_dir = os.path.dirname(os.path.realpath(__file__))
    xml_file = os.path.join(parent_dir, 'asc_to_qsch_data.xml')
    conversion_data = ET.parse(xml_file)

    # Get the root element
    root = conversion_data.getroot()

    # Get the offset and scaling
    offset = root.find('offset')
    offset_x = float(offset.get('x'))
    offset_y = float(offset.get('y'))
    scale = root.find('scaling')
    scale_x = float(scale.get('x'))
    scale_y = float(scale.get('y'))

    # Scaling the schematic
    asc_editor.scale(offset_x=offset_x, offset_y=offset_y, scale_x=scale_x, scale_y=scale_y)

    # Adding symbols to components
    # symbol_stock = {sym.find('LT_name').text: sym for sym in root.findall('component_symbols/symbol')}
    # The symbol_stock has native QSpice symbols and information on how to replace the LTSpice symbols by
    # QSpice ones. For now this is not operational
    for comp in asc_editor.components.values():
        symbol_tag = symbol_stock.get(comp.symbol, None)
        if symbol_tag is None:
            # Will try to get it from the sym folder
            print(f"Searching for symbol {comp.symbol}...")
            # TODO: this should use the default locations from AscEditor, and use search_file_in_containers, just like AscEditor does.
            for sym_root in search_paths + [
                # os.path.curdir,  # The current script directory
                os.path.split(asc_file)[0],  # The directory where the script is located
                os.path.expanduser("~/AppData/Local/LTspice/lib/sym"),
                os.path.expanduser("~/Documents/LtspiceXVII/lib/sym"),
                # os.path.expanduser(r"~\AppData\Local\Programs\ADI\LTspice\lib.zip"), # TODO: implement this
            ]:
                print(f"   {os.path.abspath(sym_root)}")
                if not os.path.exists(sym_root):  # Skipping invalid paths
                    continue
                if sym_root.endswith('.zip'):  # TODO: test if it is a file
                    pass
                    # Using an IO buffer to pass the file to the AsyEditor
                else:
                    symbol_asc_file = find_file_in_directory(sym_root, comp.symbol + '.asy')
                    if symbol_asc_file is not None:
                        print(f"Found {symbol_asc_file}")
                        symbol_asc = AsyReader(symbol_asc_file)
                        value = comp.attributes.get('Value', '<val>')
                        symbol_tag = symbol_asc.to_qsch(comp.reference, value)
                        symbol_stock[comp.symbol] = symbol_tag
                        break

        # if symbol_tree:
        #     name = symbol_tree.find("name").text
        #     offset = symbol_tree.find('LT_origin')
        #     offset_x = int(offset.get('x'))
        #     offset_y = int(offset.get('y'))
        #
        #     if comp.rotation == ERotation.R0:
        #         comp.position.X += offset_x
        #         comp.position.Y += offset_y
        #     elif comp.rotation == ERotation.R90:
        #         comp.position.X += offset_y
        #         comp.position.Y -= offset_x
        #     elif comp.rotation == ERotation.R180:
        #         comp.position.X -= offset_x
        #         comp.position.Y -= offset_y
        #     elif comp.rotation == ERotation.R270:
        #         comp.position.X -= offset_y
        #         comp.position.Y += offset_x
        #     elif comp.rotation == ERotation.M0:
        #         comp.position.X += offset_x
        #         comp.position.Y -= offset_y
        #     elif comp.rotation == ERotation.M90:
        #         comp.position.X -= offset_y
        #         comp.position.Y -= offset_x
        #     elif comp.rotation == ERotation.M180:
        #         comp.position.X -= offset_x
        #         comp.position.Y += offset_y
        #     elif comp.rotation == ERotation.M270:
        #         comp.position.X += offset_y
        #         comp.position.Y += offset_x
        #     rotation = symbol_tree.find('rotation').text
        #     if rotation == 'R0':
        #         pass  # basically do nothing
        #     elif rotation == 'R90':
        #         comp.rotation = comp.rotation + ERotation.R90
        #     elif rotation == 'R180':
        #         comp.rotation = comp.rotation + ERotation.R180
        #     elif rotation == 'R270':
        #         comp.rotation = comp.rotation + ERotation.R270
        #     elif rotation == 'M0':
        #         comp.rotation = comp.rotation.mirror_x_axis()
        #     elif rotation == 'M90':
        #         comp.rotation = comp.rotation
        #     elif rotation == 'M180':
        #         comp.rotation = comp.rotation.M180
        #     elif rotation == 'M270':
        #         comp.rotation = comp.rotation
        #
        #     # typ = symbol_tree.find("type").text
        #     # description = symbol_tree.find("description").text
        #
        #     symbol_tag = QschTag("symbol", name)
        #     # symbol_tag.items.append(QschTag("type", typ))
        #     # symbol_tag.items.append(QschTag("description:", description))
        #     for item in symbol_tree.findall('items/item'):
        #         text = item.text
        #         if item.attrib:
        #             # need to include the needed attributes
        #             fmt_dict = {}
        #             for key, value in item.attrib.items():
        #                 if key == "reference":
        #                     fmt_dict[value] = comp.reference
        #                 elif key == "value":
        #                     fmt_dict[value] = comp.attributes.get("Value", "<val>")
        #                 else:
        #                     fmt_dict[value] = comp.attributes[key]
        #             text = text.format(**fmt_dict)
        #
        #         item_tag, _ = QschTag.parse(text)
        #         symbol_tag.items.append(item_tag)
        # else:
        if comp.rotation == 90:
            comp.rotation = 270
        elif comp.rotation == 270:
            comp.rotation = 90
        elif comp.rotation == 90 + 360:
            comp.rotation = 270 + 360
        elif comp.rotation == 270 + 360:
            comp.rotation = 90 + 360

        if symbol_tag:
            comp.attributes['symbol'] = symbol_tag

    qsch_editor = QschEditor(qsch_file, create_blank=True)
    qsch_editor.copy_from(asc_editor)
    # Save the netlist
    qsch_editor.save_netlist(qsch_file)

    print("File {} converted to {}".format(asc_file, qsch_file))


if __name__ == "__main__":
    main()
    exit(0)
